import logging
import os
from typing import Dict, List, Union

import numpy as np
from qdrant_client import QdrantClient
from qdrant_client.http.models import SearchRequest
from qdrant_client.models import Distance, PointStruct, VectorParams

from geniml.const import PKG_NAME
from geniml.search.const import (
    DEFAULT_COLLECTION_NAME,
    DEFAULT_DIM,
    DEFAULT_QDRANT_DIST,
    DEFAULT_QDRANT_HOST,
    DEFAULT_QDRANT_PORT,
    DEFAULT_QUANTIZATION_CONFIG,
)

from ..utils import verify_load_inputs
from .abstract import EmSearchBackend

_LOGGER = logging.getLogger(PKG_NAME)


def queries_to_requests(
    queries: np.ndarray,
    limit: int,
    with_payload: bool = True,
    with_vectors: bool = True,
    offset: int = 0,
) -> List[SearchRequest]:
    """
    Prepare all search requests for each query vector in a batch

    :param queries: see docstring of QdrantBackend.batch_search
    :param limit:
    :param with_payload:
    :param with_vectors:
    :param offset:
    """
    requests = []
    for query in queries:
        if query.ndim > 1:
            # that each request is from one single query vector
            requests.extend(queries_to_requests(query, limit, with_payload, with_vectors, offset))
        else:
            requests.append(
                SearchRequest(
                    vector=query,
                    limit=limit,
                    with_vector=with_vectors,
                    with_payload=with_payload,
                    offset=offset,
                )
            )
    return requests


def results_processing(search_results, with_payload: bool, with_vectors: bool) -> List[Dict]:
    """
    Process the search result into unified format: list of dictionaries

    :param search_results: result of qdrant client similarity search
    :type search_results: search result of qdrant client
    :param with_payload: see docstring of QdrantBackend.search
    :param with_vectors:
    """
    output_list = []
    for result in search_results:
        # build each dictionary
        result_dict = {"id": result.id, "score": result.score}
        if with_payload:
            result_dict["payload"] = result.payload
        if with_vectors:
            result_dict["vector"] = result.vector
        output_list.append(result_dict)
    return output_list


class QdrantBackend(EmSearchBackend):
    """A search backend that uses a qdrant server to store and search embeddings"""

    def __init__(
        self,
        dim: int = DEFAULT_DIM,
        dist: Distance = DEFAULT_QDRANT_DIST,
        collection: str = DEFAULT_COLLECTION_NAME,
        qdrant_host: str = DEFAULT_QDRANT_HOST,
        qdrant_port: int = DEFAULT_QDRANT_PORT,
        qdrant_api_key: str = None,
    ):
        """
        Connect to Qdrant on commandline first:
        (Ubuntu Linux terminal)
        sudo docker run -p 6333:6333     -v $(pwd)/qdrant_storage:/qdrant/storage     qdrant/qdrant

        :param config: the vector parameter
        :param collection: name of collection
        :param qdrant_host: host of qdrant server
        :param qdrant_port: port of qdrant server
        :param qdrant_api_key: api key
        """
        super().__init__()
        self.collection = collection
        self.config = VectorParams(size=dim, distance=dist)
        self.url = os.environ.get("QDRANT_HOST", qdrant_host)
        self.port = os.environ.get("QDRANT_PORT", qdrant_port)
        self.qd_client = QdrantClient(
            url=self.url,
            port=self.port,
            api_key=os.environ.get("QDRANT_API_KEY", qdrant_api_key),
        )

        # Create collection only if it does not exist
        try:
            collection_info = self.qd_client.get_collection(collection_name=self.collection)
            _LOGGER.info(
                f"Using collection {self.collection} with {collection_info.points_count} points."
            )
        except Exception:  # qdrant_client.http.exceptions.UnexpectedResponse
            _LOGGER.info(f"Collection {self.collection} does not exist, creating it.")
            self.qd_client.recreate_collection(
                collection_name=self.collection,
                vectors_config=self.config,
                quantization_config=DEFAULT_QUANTIZATION_CONFIG,
            )

    def load(
        self,
        vectors: np.ndarray,
        ids: Union[List[str], None] = None,
        payloads: Union[List[Dict[str, str]], None] = None,
    ):
        """
        Upload vectors and their labels into qdrant storage.

        :param vectors: embedding vectors, a np.ndarray with shape of (n, <vector size>)
        :param ids: list of n point ids, or None to generate ids automatically
        :param payloads: optional list of n dictionaries that contain vector metadata
        :return:
        """

        if not ids:
            start = len(self)
            ids = list(range(start, start + len(payloads)))

        verify_load_inputs(vectors, ids, payloads)

        points = [
            PointStruct(id=ids[i], vector=vectors[i].tolist(), payload=payloads[i])
            for i in range(len(payloads))
        ]
        self.qd_client.upsert(
            collection_name=self.collection,
            points=points,
        )

    def search(
        self,
        query: np.ndarray,
        limit: int,
        with_payload: bool = True,
        with_vectors: bool = True,
        offset: int = 0,
    ) -> Union[
        List[Dict[str, Union[int, float, Dict[str, str], List[float]]]],
        List[List[Dict[str, Union[int, float, Dict[str, str], List[float]]]]],
    ]:
        """
         with a given query vector, get k nearest neighbors from vectors in the collection

        :param query: a vector to search
        :param limit: number of returned results
        :param with_payload: whether payload is included in the result
        :param with_vectors: whether the stored vector is included in the result
        :param offset: the offset of the search results
        :return: a list of dictionary that contains the search results in this format:
        {
            "id": <id>
            "score": <score>
            "payload": {
                <information of the vector>
            }
            "vector": [<the vector>]
        }
        """
        if query.ndim > 1:
            return self.batch_search(query, limit, with_payload, with_vectors, offset)
        # KNN search in qdrant client
        search_results = self.qd_client.search(
            collection_name=self.collection,
            query_vector=query,
            limit=limit,
            with_payload=with_payload,
            with_vectors=with_vectors,
            offset=offset,
        )

        # add the results in to the output list
        return results_processing(search_results, with_payload, with_vectors)

    def batch_search(
        self,
        queries: np.ndarray,
        limit: int,
        with_payload: bool = True,
        with_vectors: bool = True,
        offset: int = 0,
    ) -> List[List[Dict[str, Union[int, float, Dict[str, str], List[float]]]]]:
        """

        :param queries: multiple search vectors, np.ndarray with shape of (n, dim)
        :param limit: see docstring of def search
        :type limit:
        :param with_payload:
        :param with_vectors:
        :param offset:
        :return: results of all search requests with each vector in queries
        """
        output_list = []
        # build all search requests
        requests = queries_to_requests(queries, limit, with_payload, with_vectors, offset)

        search_results = self.qd_client.search_batch(
            collection_name=self.collection, requests=requests
        )

        # add the results in to the output list
        for batch in search_results:
            batch_list = results_processing(batch, with_payload, with_vectors)
            output_list.append(batch_list)
        return output_list

    def __len__(self) -> int:
        """
        Return the number of embeddings in the backend
        """
        return self.qd_client.get_collection(collection_name=self.collection).vectors_count

    def retrieve_info(
        self, ids: Union[List[int], int, List[str], str], with_vectors: bool = False
    ) -> Union[
        Dict[str, Union[int, str, List[float], Dict[str, str]]],
        List[Dict[str, Union[int, str, List[float], Dict[str, str]]]],
    ]:
        """
        With a given list of storage ids, return the information of these vectors

        :param ids: list of ids, or a single id
        :param with_vectors:  whether the vectors themselves will also be returned in the output
        :return: if ids is one id, a dictionary similar to the output of search() will be returned, without "score";
        if ids is a list, a list of dictionaries will be returned
        """
        if not isinstance(ids, list):
            # retrieve() only takes iterable input
            ids = [ids]

        # add hyphen to uuid if missing
        for i in range(len(ids)):
            id_ = ids[i]
            if isinstance(id_, str):
                if not "-" in id_:
                    ids[i] = f"{id_[:8]}-{id_[8:12]}-{id_[12:16]}-{id_[16:20]}-{id_[20:]}"

        output_list = []
        retrievals = self.qd_client.retrieve(
            collection_name=self.collection,
            ids=ids,
            with_payload=True,
            with_vectors=with_vectors,  # no need vectors
        )

        retrieval_dict = {result.id: result for result in retrievals}

        # retrieve() of qd client does not return result in the order of ids in the list
        # get the retrieval result in output by id order
        for id_ in ids:
            try:
                result = retrieval_dict[id_]
            except:
                _LOGGER.warning(f"Warning: no id stored in backend matches {id_}.")
                continue
            result_dict = {"id": result.id, "payload": result.payload}
            if with_vectors:
                result_dict["vector"] = result.vector
            output_list.append(result_dict)

        # with just one id, only the dictionary instead of the list will be returned
        if len(output_list) == 1:
            return output_list[0]
        else:
            return output_list

    def __str__(self):
        n_items = len(self)
        msg = f"""QdrantBackend
            n items: {n_items}
            url: {self.url}:{self.port},
            collection: {self.collection}
            """
        return msg

    def __repr__(self):
        n_items = len(self)
        msg = f"""QdrantBackend
            n items: {n_items}
            url: {self.url}:{self.port},
            collection: {self.collection}
            """
        return msg



from enum import Enum
from pathlib import Path

PKG_NAME = "bedspace"

CACHE_DIR = MODEL_CACHE_DIR = str(Path.home() / f".{PKG_NAME}")

PREPROCESS_CMD = "preprocess"
TRAIN_CMD = "train"
DISTANCES_CMD = "distances"
SEARCH_CMD = "search"


DEFAULT_NUM_SEARCH_RESULTS = 10
DEFAULT_NUM_EPOCHS = 50
DEFAULT_DIM = 100
DEFAULT_LEARNING_RATE = 0.05
DEFAULT_THRESHOLD = 0.5


# search type enum. Can be "l2r", "r2l", or "r2r"
class SearchType(Enum):
    l2r = "l2r"
    r2l = "r2l"
    r2r = "r2r"



from abc import ABC, abstractmethod
from typing import Dict, List, Union

import numpy as np

from ...io import RegionSet
from ..backends.abstract import EmSearchBackend
from ..query2vec.abstract import Query2Vec


class BEDSearchInterface(ABC):
    """
    An abstract class representing BED files search interface.
    The query will be embedded by one of the subclass of Query2Vec,
    and the embedding is used to do KNN search in the backend
    where BED embeddings are stored.
    """

    def __init__(self, backend: EmSearchBackend, query2vec: Query2Vec) -> None:
        self.backend = backend
        self.query2vec = query2vec

    @abstractmethod
    def query_search(
        self,
        query: Union[str, RegionSet, np.ndarray],
        limit: int,
        with_payload: bool = True,
        with_vectors: bool = True,
        offset: int = 0,
    ) -> List[Dict]:
        raise NotImplementedError



try:
    import torch
    import torch.nn as nn
except ImportError:
    raise ImportError(
        "Please install Machine Learning dependencies by running 'pip install geniml[ml]'"
    )

from .const import DEFAULT_EMBEDDING_DIM, POOLING_TYPES


class Word2Vec(nn.Module):
    """
    Word2Vec model.
    """

    def __init__(
        self,
        vocab_size: int,
        embedding_dim: int = DEFAULT_EMBEDDING_DIM,
    ):
        super().__init__()
        self.vocab_size = vocab_size
        self.embedding_dim = embedding_dim
        self.projection = nn.Embedding(vocab_size, embedding_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.projection(x)
        return x


class Region2Vec(Word2Vec):
    def __init__(
        self,
        vocab_size: int,
        embedding_dim: int = DEFAULT_EMBEDDING_DIM,
    ):
        super().__init__(vocab_size, embedding_dim)


class MeanPooling(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.mean(x, dim=1)


class MaxPooling(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.max(x, dim=1)


class RegionSet2Vec(nn.Module):
    def __init__(self, region2vec: Region2Vec, pooling: POOLING_TYPES = "mean"):
        """
        Initialize the RegionSet2Vec. RegionSet2Vec is a wrapper around the Region2Vec model that allows
        pooling over a set of regions. This is useful for classification tasks where the input is a set of
        regions, such as classifying a cell type based on the set of regions that are accessible.

        :param Union[Region2Vec, str] region2vec: Either a Region2Vec instance or a path to a huggingface model.
        :param POOLING_TYPES pooling: The pooling type to use. Either "mean" or "max".
        """
        super().__init__()

        self.region2vec: Region2Vec = region2vec
        # assign the pooling layer based on mean or max
        if pooling == "mean":
            self.pooling = MeanPooling()
        elif pooling == "max":
            self.pooling = MaxPooling()
        else:
            raise ValueError(f"Invalid pooling type {pooling} passed.")

    def forward(self, x) -> torch.Tensor:
        x = self.region2vec(x)
        x = self.pooling(x)
        return x



# metadata from csv
DEFAULT_GENOME_KEY = "sample_genome"
DEFAULT_SERIES_KEY = "gse"
DEFAULT_FILE_KEY = "file"
BIO_GPT_REPO = "microsoft/biogpt"
BIO_BERT_REPO = "dmis-lab/biobert-v1.1"


DEFAULT_TRAIN_P = 0.85 * 0.9
DEFAULT_VALIDATE_P = 0.85 * 0.1

# sentence transformer model from hugging face
DEFAULT_NL_EMBEDDING_MODEL = "sentence-transformers/all-MiniLM-L12-v2"
DEFAULT_MAX_SEQ_LENGTH = 1000

DEFAULT_NUM_EPOCHS = 1000
DEFAULT_NUM_UNITS = 256
DEFAULT_BATCH_SIZE = 1
DEFAULT_OPTIMIZER_NAME = "Adam"
DEFAULT_LOSS_NAME = "cosine_embedding_loss"
DEFAULT_MARGIN = 0.0
# embedding dimension of Region2Vec: https://huggingface.co/databio/r2v-ChIP-atlas-hg38
DEFAULT_EMBEDDING_DIM = (100,)
# default learning rate of Adam optimizer
DEFAULT_LEARNING_RATE = 0.001

# if validation loss does not improve after patience*epoches, training stops
DEFAULT_PATIENCE = 0.2


DEFAULT_DATALOADER_SHUFFLE = True
MODULE_NAME = "text2bednn"
CONFIG_FILE_NAME = "config.yaml"
TORCH_MODEL_FILE_NAME_PATTERN = "v2v_{callback}_{checkpoint}.pt"
DEFAULT_MUST_TRAINED = True
DEFAULT_PLOT_FILE_NAME = "training_history"
DEFAULT_PLOT_TITLE = "Diagram of loss and epochs"
DEFAULT_HUGGINGFACE_MODEL_NAME = "checkpoint.pt"



import argparse
import os


def get_file_rows(file_path: str) -> int:
    """Counts how many files are included.

    Args:
        file_path (str): The path to a file.

    Returns:
        int: The number of rows in the file.
    """
    count = 0
    with open(file_path, "r") as f:
        for line in f:
            count += 1
    return count


def split_file(file_path: str, dest_folder: str, num_parts: int) -> None:
    """Splits a list of files into num_parts non-overlapping batches.

    This is a helper function for tokenization in parallel. The dest_folder
    must be empty.

    Args:
        file_path (str): The path to a file with BED file names per row.
        dest_folder (str): The folder to store split file lists.
        num_parts (int): Number of batches to split.
    """
    if os.path.exists(dest_folder):
        print("Folder exists")
        return
    os.makedirs(dest_folder)
    count = get_file_rows(file_path)

    list_arr = []
    with open(file_path, "r") as f:
        for line in f:
            list_arr.append(line.strip())

    num_per_file = int(count / num_parts)
    if num_per_file == 0:
        print("No need to split")
        return
    else:
        last_file_num = count - num_per_file * (num_parts - 1)
        num_arr = [num_per_file] * (num_parts - 1) + [last_file_num]
        pos = 0
        for index in range(num_parts):
            fname = os.path.join(dest_folder, f"split_{index}.txt")
            with open(fname, "w") as fout:
                for _ in range(num_arr[index]):
                    fout.write(list_arr[pos])
                    fout.write("\n")
                    pos = pos + 1
        assert count == pos, "Missing some files"


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--file-path", default="", help="path to a file list")
    parser.add_argument("--dest-folder", default="", help="where to store the split files")
    parser.add_argument(
        "--num-parts",
        type=int,
        default=5,
        help="split the original file list to the specified parts",
    )
    args = parser.parse_args()

    split_file(args.file_path, args.dest_folder, args.num_parts)



import os

import numpy as np
import pytest
import torch

from geniml.io.io import Region, RegionSet
from geniml.region2vec.main import Region2Vec, Region2VecExModel
from geniml.region2vec.utils import Region2VecDataset
from geniml.tokenization.main import TreeTokenizer


@pytest.fixture
def universe_file():
    return "tests/data/universe.uniq.bed"


def test_init_region2vec():
    model = Region2Vec(
        vocab_size=10000,
        embedding_dim=100,
    )
    assert model is not None


def test_make_region2vec_dataset():
    path_to_data = "tests/data/gtok_sample/"
    dataset = Region2VecDataset(path_to_data)

    first = next(iter(dataset))
    assert all([isinstance(x, int) for x in first])


@pytest.mark.skip(reason="Model is too big to download in the runner, takes too long.")
def test_pretrained_model():
    model = Region2VecExModel("databio/r2v-ChIP-atlas-hg38-v2")

    region = Region("chr1", 63403166, 63403785)
    embedding = model.encode(region)

    assert embedding is not None
    assert isinstance(embedding, np.ndarray)


def test_r2v_pytorch_forward():
    vocab_size = 10000
    embedding_dim = 100

    model = Region2Vec(vocab_size, embedding_dim)
    assert model is not None

    # create a random tensor with 10 tokens
    x = torch.randint(low=0, high=100, size=(10,))
    y = model.forward(x)
    assert y.shape == (10, 100)


def test_r2v_pytorch_tokenizer_is_file_on_disk(universe_file: str):
    model = Region2VecExModel(tokenizer=universe_file)
    assert model is not None
    assert (
        len(model.tokenizer) == 2_385
    )  # 2378 + 7 special tokens (unk, pad, mask, sep, cls, eos, bos)


@pytest.mark.skip(reason="Downloading the model takes too long.")
def test_r2v_pytorch_tokenizer_is_on_hf():
    model = Region2VecExModel(tokenizer="databio/r2v-ChIP-atlas-hg38-v2")
    assert model is not None
    assert len(model.tokenizer) == 1_698_713


def test_r2v_pytorch_exmodel_train(universe_file: str):
    model = Region2VecExModel(tokenizer=TreeTokenizer(universe_file))
    assert model is not None

    dataset = Region2VecDataset("tests/data/gtok_sample/", convert_to_str=True)

    loss = model.train(dataset, epochs=10, min_count=1)
    assert loss


def test_r2v_pytorch_encode(universe_file: str):
    model = Region2VecExModel(tokenizer=TreeTokenizer(universe_file))
    assert model is not None

    r = Region("chr1", 63403166, 63403785)
    embedding = model.encode(r)
    assert embedding is not None
    assert isinstance(embedding, np.ndarray)
    assert embedding.shape == (1, 100)

    rs = RegionSet("tests/data/to_tokenize.bed")
    embedding = model.encode(rs)
    assert embedding is not None
    assert isinstance(embedding, np.ndarray)
    assert embedding.shape == (13, 100)


def test_save_load_pytorch_exmodel(universe_file: str):
    model = Region2VecExModel(tokenizer=TreeTokenizer(universe_file))
    assert model is not None

    dataset = Region2VecDataset("tests/data/gtok_sample/", convert_to_str=True)
    loss = model.train(dataset, epochs=10, min_count=1)

    before_embedding = model.encode(Region("chr1", 63403166, 63403785))
    assert loss
    try:
        # save the model
        model.export("tests/data/test_model/")
        assert os.path.exists("tests/data/test_model/checkpoint.pt")
        assert os.path.exists("tests/data/test_model/universe.bed")

        # load in
        model_loaded = Region2VecExModel.from_pretrained("tests/data/test_model")

        # the region embeddings should be the same
        after_embedding = model_loaded.encode(Region("chr1", 63403166, 63403785))
        assert np.allclose(before_embedding, after_embedding)

    finally:
        try:
            os.remove("tests/data/test_model/checkpoint.pt")
            os.remove("tests/data/test_model/universe.bed")
            os.remove("tests/data/test_model/config.yaml")
            os.rmdir("tests/data/test_model/")
        except Exception as e:
            # just try to remove it, if it doesn't work, then pass, means something
            # else wrong occurred up the stack
            print(e)
            pass



import os
from logging import getLogger
from typing import Union

import numpy as np
import scanpy as sc
import torch
from huggingface_hub import hf_hub_download
from rich.progress import track

from ..region2vec.const import (
    CONFIG_FILE_NAME,
    DEFAULT_EMBEDDING_DIM,
    DEFAULT_EPOCHS,
    DEFAULT_MIN_COUNT,
    DEFAULT_WINDOW_SIZE,
    MODEL_FILE_NAME,
    POOLING_METHOD_KEY,
    POOLING_TYPES,
    UNIVERSE_FILE_NAME,
)
from ..region2vec.main import Region2Vec
from ..region2vec.utils import (
    Region2VecDataset,
    export_region2vec_model,
    load_local_region2vec_model,
    train_region2vec_model,
)
from ..tokenization.main import AnnDataTokenizer, Tokenizer
from .const import MODULE_NAME

_GENSIM_LOGGER = getLogger("gensim")
_LOGGER = getLogger(MODULE_NAME)

# demote gensim logger to warning
_GENSIM_LOGGER.setLevel("WARNING")


class ScEmbed:
    def __init__(
        self,
        model_path: str = None,
        tokenizer: AnnDataTokenizer = None,
        device: str = None,
        pooling_method: POOLING_TYPES = "mean",
        **kwargs,
    ):
        """
        Initialize ScEmbed.

        :param str model_path: Path to the pre-trained model on huggingface.
        :param embedding_dim: Dimension of the embedding.
        :param kwargs: Additional keyword arguments to pass to the model.
        """
        super().__init__()
        self.model_path: str = model_path
        self.tokenizer: AnnDataTokenizer
        self.trained: bool = False
        self._model: Region2Vec = None
        self.pooling_method: POOLING_TYPES = pooling_method

        if model_path is not None:
            self._init_from_huggingface(model_path)
            self.trained = True

        elif tokenizer is not None:
            self._init_model(tokenizer, **kwargs)

        # set the device
        self._target_device = torch.device(
            device if device else ("cuda" if torch.cuda.is_available() else "cpu")
        )

    def _init_tokenizer(self, tokenizer: Union[AnnDataTokenizer, str]):
        """
        Initialize the tokenizer.

        :param tokenizer: Tokenizer to add to the model.
        """
        if isinstance(tokenizer, str):
            if os.path.exists(tokenizer):
                self.tokenizer = AnnDataTokenizer(tokenizer)
            else:
                self.tokenizer = AnnDataTokenizer.from_pretrained(
                    tokenizer
                )  # download from huggingface (or at least try to)
        elif isinstance(tokenizer, AnnDataTokenizer):
            self.tokenizer = tokenizer
        else:
            raise TypeError("tokenizer must be of type AnnDataTokenizer or str.")

    def _init_model(self, tokenizer, **kwargs):
        """
        Initialize the core model. This will initialize the model from scratch.

        :param kwargs: Additional keyword arguments to pass to the model.
        """
        self._init_tokenizer(tokenizer)

        self._model = Region2Vec(
            len(self.tokenizer),
            embedding_dim=kwargs.get("embedding_dim", DEFAULT_EMBEDDING_DIM),
        )

    @property
    def model(self):
        """
        Get the core Region2Vec model.
        """
        return self._model

    def add_tokenizer(self, tokenizer: Tokenizer, **kwargs):
        """
        Add a tokenizer to the model. This should be use when the model
        is not initialized with a tokenizer.

        :param tokenizer: Tokenizer to add to the model.
        :param kwargs: Additional keyword arguments to pass to the model.
        """
        if self._model is not None:
            raise RuntimeError("Cannot add a tokenizer to a model that is already initialized.")

        self.tokenizer = tokenizer
        if not self.trained:
            self._init_model(**kwargs)

    def _load_local_model(self, model_path: str, vocab_path: str, config_path: str):
        """
        Load the model from a checkpoint.

        :param str model_path: Path to the model checkpoint.
        :param str vocab_path: Path to the vocabulary file.
        """
        _model, config = load_local_region2vec_model(model_path, config_path)
        tokenizer = AnnDataTokenizer(vocab_path, verbose=True)

        self._model = _model
        self.tokenizer = tokenizer

        if POOLING_METHOD_KEY in config:
            self.pooling_method = config[POOLING_METHOD_KEY]

    def _init_from_huggingface(
        self,
        model_path: str,
        model_file_name: str = MODEL_FILE_NAME,
        universe_file_name: str = UNIVERSE_FILE_NAME,
        config_file_name: str = CONFIG_FILE_NAME,
        **kwargs,
    ):
        """
        Initialize the model from a huggingface model. This uses the model path
        to download the necessary files and then "build itself up" from those. This
        includes both the actual model and the tokenizer.

        :param str model_path: Path to the pre-trained model on huggingface.
        :param str model_file_name: Name of the model file.
        :param str universe_file_name: Name of the universe file.
        :param kwargs: Additional keyword arguments to pass to the hf download function.
        """
        model_file_path = hf_hub_download(model_path, model_file_name, **kwargs)
        universe_path = hf_hub_download(model_path, universe_file_name, **kwargs)
        config_path = hf_hub_download(model_path, config_file_name, **kwargs)

        self._load_local_model(model_file_path, universe_path, config_path)

    @classmethod
    def from_pretrained(
        cls,
        path_to_files: str,
        model_file_name: str = MODEL_FILE_NAME,
        universe_file_name: str = UNIVERSE_FILE_NAME,
        config_file_name: str = CONFIG_FILE_NAME,
    ) -> "ScEmbed":
        """
        Load the model from a set of files that were exported using the export function.

        :param str path_to_files: Path to the directory containing the files.
        :param str model_file_name: Name of the model file.
        :param str universe_file_name: Name of the universe file.
        :param str config_file_name: Name of the config file.

        :return: The loaded model.
        """
        model_file_path = os.path.join(path_to_files, model_file_name)
        universe_file_path = os.path.join(path_to_files, universe_file_name)
        config_file_path = os.path.join(path_to_files, config_file_name)

        instance = cls()
        instance._load_local_model(model_file_path, universe_file_path, config_file_path)
        instance.trained = True

        return instance

    def train(
        self,
        dataset: Region2VecDataset,
        window_size: int = DEFAULT_WINDOW_SIZE,
        epochs: int = DEFAULT_EPOCHS,
        min_count: int = DEFAULT_MIN_COUNT,
        num_cpus: int = 1,
        seed: int = 42,
        save_checkpoint_path: str = None,
        gensim_params: dict = {},
        load_from_checkpoint: str = None,
    ) -> bool:
        """
        Train the model.

        :param Region2VecDataset data: Data to train on. This is a dataset of tokens.
        :param int window_size: Window size for the model.
        :param int epochs: Number of epochs to train for.
        :param int min_count: Minimum count for a region to be included in the vocabulary.
        :param int num_cpus: Number of cpus to use for training.
        :param int seed: Seed to use for training.
        :param str save_checkpoint_path: Path to save the model checkpoints to.
        :param dict gensim_params: Additional parameters to pass to the gensim model.
        :param str load_from_checkpoint: Path to a checkpoint to load from.

        :return bool: Whether or not the model was trained.
        """

        # validate a model exists
        if self._model is None:
            raise RuntimeError(
                "Cannot train a model that has not been initialized. Please initialize the model first using a tokenizer or from a huggingface model."
            )

        gensim_model = train_region2vec_model(
            dataset,
            embedding_dim=self._model.embedding_dim,
            window_size=window_size,
            epochs=epochs,
            min_count=min_count,
            num_cpus=num_cpus,
            seed=seed,
            save_checkpoint_path=save_checkpoint_path,
            gensim_params=gensim_params,
            load_from_checkpoint=load_from_checkpoint,
        )

        # once done training, set the weights of the pytorch model in self._model
        for id in track(
            gensim_model.wv.key_to_index,
            total=len(gensim_model.wv.key_to_index),
            description="Setting weights.",
        ):
            self._model.projection.weight.data[int(id)] = torch.tensor(gensim_model.wv[id])

        # set the model as trained
        self.trained = True

        return True

    def export(
        self,
        path: str,
        checkpoint_file: str = MODEL_FILE_NAME,
        universe_file: str = UNIVERSE_FILE_NAME,
        config_file: str = CONFIG_FILE_NAME,
    ):
        """
        Function to facilitate exporting the model in a way that can
        be directly uploaded to huggingface. This exports the model
        weights and the vocabulary.

        :param str path: Path to export the model to.
        """
        # make sure the model is trained
        if not self.trained:
            raise RuntimeError("Cannot export an untrained model.")

        export_region2vec_model(
            self._model,
            self.tokenizer,
            path,
            checkpoint_file=checkpoint_file,
            universe_file=universe_file,
            config_file=config_file,
        )

    def encode(self, regions: Union[sc.AnnData, str], pooling: POOLING_TYPES = None) -> np.ndarray:
        """
        Get the vector for a region.

        :param Region region: Region to get the vector for.
        :param str pooling: Pooling type to use.

        :return np.ndarray: Vector for the region.
        """
        # allow the user to override the pooling method
        pooling = pooling or self.pooling_method

        # data validation
        if not (isinstance(regions, sc.AnnData) or isinstance(regions, str)):
            raise TypeError(
                f"Regions must be of type AnnData or str, not {type(regions).__name__}"
            )
        if isinstance(regions, str):
            regions = sc.read_h5ad(regions)

        if pooling not in ["mean", "max"]:
            raise ValueError(f"pooling must be one of {POOLING_TYPES}")

        # tokenize the region
        tokens = self.tokenizer(regions)
        tokens = [[t.id for t in sublist] for sublist in tokens]

        # get the vector
        embeddings = []
        for token_set in track(tokens, total=len(tokens), description="Getting embeddings"):
            region_embeddings = self._model.projection(torch.tensor(token_set))
            if pooling == "mean":
                region_embeddings = torch.mean(region_embeddings, axis=0).detach().numpy()
            elif pooling == "max":
                region_embeddings = torch.max(region_embeddings, axis=0).values.detach().numpy()
            else:
                # this should be unreachable
                raise ValueError(f"pooling must be one of {POOLING_TYPES}")
            embeddings.append(region_embeddings)

        return np.vstack(embeddings)



import os

import botocore
import genomicranges
import pytest

from geniml.bbclient import BBClient
from geniml.exceptions import TokenizedFileNotFoundInCacheError
from geniml.io import BedSet, RegionSet

DATA_TEST_FOLDER = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "tests",
    "data",
    "io_data",
    "bed",
)
ALL_BEDFILE_PATH = [os.path.join(DATA_TEST_FOLDER, x) for x in os.listdir(DATA_TEST_FOLDER)]


@pytest.fixture
def cache_path():
    return "./data/geniml_bb_cache"


@pytest.fixture
def bedset_id():
    """
    identifier of bedset
    """
    return "2f5fbf7a65c1b666e38bccf7ed9da5e6"


@pytest.fixture
def bedfile_id():
    return "38856b21ff44f584e48081e0db51db0c"


@pytest.fixture
def local_bedfile_path():
    return ALL_BEDFILE_PATH[0]


@pytest.fixture
def local_bedfile_list():
    return ALL_BEDFILE_PATH


class TestBedCaching:
    pass


class TestBBClientCaching:
    def test_init(self, cache_path):
        """
        Test initialization of BBClient
        """
        bbclient = BBClient(cache_folder=cache_path)
        assert isinstance(bbclient, BBClient)

    def test_init_no_cache_folder(self):
        """
        Test initialization of BBClient without cache folder
        """
        with pytest.raises(TypeError):
            BBClient(cache_folder=None)

    @pytest.mark.parametrize("bedfile_path", ALL_BEDFILE_PATH)
    def test_bed_caching_from_path(self, bedfile_path, tmp_path):
        bbclient = BBClient(cache_folder=tmp_path)
        bedfile_id = bbclient.add_bed_to_cache(bedfile_path)
        assert bedfile_id is not None

    @pytest.mark.parametrize("bedfile_path", ALL_BEDFILE_PATH)
    def test_bed_caching_from_region_set(self, tmp_path, bedfile_path):
        bbclient = BBClient(cache_folder=tmp_path)
        bedfile = RegionSet(bedfile_path)
        bbclient.add_bed_to_cache(bedfile)
        path_in_cache = bbclient.seek(bedfile.identifier)
        assert bedfile.compute_bed_identifier() == os.path.split(path_in_cache)[1].split(".")[0]

    def test_bedset_caching(self, tmp_path, local_bedfile_list):
        bbclient = BBClient(cache_folder=tmp_path)
        bedset = BedSet(local_bedfile_list)
        bedset_id = bbclient.add_bedset_to_cache(bedset)
        path_in_cache = bbclient.seek(bedset_id)
        assert bedset_id == os.path.split(path_in_cache)[1].split(".")[0]

    def test_bed_not_in_cache_error(self, tmp_path, local_bedfile_path):
        bbclient = BBClient(cache_folder=tmp_path)
        # skip caching
        bedfile = RegionSet(local_bedfile_path)
        with pytest.raises(FileNotFoundError):
            bbclient.seek(bedfile.identifier)

    def test_bedset_not_in_cache_error(self, tmp_path, local_bedfile_list):
        bbclient = BBClient(cache_folder=tmp_path)
        bedset_id = BedSet(local_bedfile_list).identifier
        with pytest.raises(FileNotFoundError):
            bbclient.seek(bedset_id)

    def test_remove_bed_from_cache(self, tmp_path, local_bedfile_path):
        bbclient = BBClient(cache_folder=tmp_path)
        bedfile_id = bbclient.add_bed_to_cache(local_bedfile_path)
        assert bbclient.seek(bedfile_id)
        bbclient.remove_bedfile_from_cache(bedfile_id)
        with pytest.raises(FileNotFoundError):
            bbclient.seek(bedfile_id)

    def test_remove_bedset_from_cache(self, tmp_path, local_bedfile_list):
        bbclient = BBClient(cache_folder=tmp_path)
        bedset = BedSet(local_bedfile_list)
        bedset_id = bbclient.add_bedset_to_cache(bedset)
        assert bbclient.seek(bedset_id)
        bbclient.remove_bedset_from_cache(bedset_id)
        with pytest.raises(FileNotFoundError):
            bbclient.seek(bedset_id)

    @pytest.mark.parametrize("bedfile_path", ALL_BEDFILE_PATH)
    def test_bioc_cache_bedfile(self, bedfile_path, tmp_path):
        bbclient = BBClient(cache_folder=tmp_path)
        bedfile_id = bbclient.add_bed_to_cache(bedfile_path)
        assert bbclient._bedfile_cache.get(bedfile_id).fpath == bbclient.seek(bedfile_id)

    def test_bioc_cache_bedset(self, tmp_path, local_bedfile_list):
        bbclient = BBClient(cache_folder=tmp_path)
        bedset = BedSet(local_bedfile_list)
        bedset_id = bbclient.add_bedset_to_cache(bedset)
        path_in_cache = bbclient.seek(bedset_id)
        assert bbclient._bedset_cache.get(bedset_id).fpath == path_in_cache


class TestBBClientTokens:
    def test_add_token_to_cache(self, tmp_path):
        bbclient = BBClient(cache_folder=tmp_path)
        bbclient.cache_tokens("test_token", "test_id", [1, 2, 5, 7])

        assert list(bbclient.load_bed_tokens("test_token", "test_id")) == [1, 2, 5, 7]

        bbclient.remove_tokens("test_token", "test_id")

        with pytest.raises(TokenizedFileNotFoundInCacheError):
            bbclient.load_bed_tokens("test_token", "test_id")


class TestS3Caching:
    def test_upload_s3(self, mocker, local_bedfile_path, tmp_path):
        bbclient = BBClient(cache_folder=tmp_path)
        bedfile_id = bbclient.add_bed_to_cache(local_bedfile_path)
        upload_mock = mocker.patch(
            "boto3.s3.inject.upload_file",
        )
        bbclient.add_bed_to_s3(bedfile_id, s3_path="test_test")
        assert upload_mock.called

    def test_download_s3(self, mocker, local_bedfile_path, tmp_path):
        bbclient = BBClient(cache_folder=tmp_path)
        download_mock = mocker.patch(
            "boto3.s3.inject.download_file",
        )
        bbclient.get_bed_from_s3("test_id", s3_path="test_test")
        assert download_mock.called

    def test_download_s3_404(self, mocker, local_bedfile_path, tmp_path):
        bbclient = BBClient(cache_folder=tmp_path)
        download_mock = mocker.patch(
            "boto3.s3.inject.download_file",
            side_effect=botocore.exceptions.ClientError(
                {"Error": {"Code": "404"}}, "operation_name"
            ),
        )
        with pytest.raises(FileNotFoundError):
            bbclient.get_bed_from_s3("test_id", s3_path="test_test")

        assert download_mock.called


# TODO: rewrite it so that it makes the requests
# @pytest.mark.bedbase
@pytest.mark.skipif(
    "not config.getoption('--bedbase')",
    reason="Only run when --bedbase is given",
)
def test_bedbase_caching(tmp_path, bedset_id, bedfile_id, request):
    """
    Testing caching BED files and BED sets from bedbase files
    only tested when
    """
    bbclient = BBClient(cache_folder=tmp_path)
    bedset = bbclient.load_bedset(bedset_id)
    # check the GenomicRangesList from the BED set
    grl = bedset.to_granges_list()
    assert isinstance(grl, genomicranges.GenomicRangesList)

    # check the path and identifier of BED files which the loaded BED set contains
    for bedfile in bedset:
        bedfile_path = bbclient.seek(bedfile.compute_bed_identifier())
        print(bedfile.compute_bed_identifier())
        print(os.path.split(bedfile_path)[1].split(".")[0])
        assert bedfile.compute_bed_identifier() == os.path.split(bedfile_path)[1].split(".")[0]

    bedfile = bbclient.load_bed(bedfile_id)

    # check the GenomicRanges from the BED file
    gr = bedfile.to_granges()
    assert isinstance(gr, genomicranges.GenomicRanges)

    # check the path and identifier of BED file
    bedfile_path = bbclient.seek(bedfile_id)
    assert bedfile.compute_bed_identifier() == os.path.split(bedfile_path)[1].split(".")[0]



from typing import Dict, List, Union

import numpy as np

from ...io import RegionSet
from ..backends import HNSWBackend, QdrantBackend
from ..query2vec import BED2Vec
from .abstract import BEDSearchInterface


class BED2BEDSearchInterface(BEDSearchInterface):
    """Search interface for the query that is a region set"""

    def __init__(
        self,
        backend: Union[QdrantBackend, HNSWBackend],
        query2vec: Union[str, BED2Vec],
    ):
        """

        :param backend: the backend where vectors are stored

        :param query2vec: a BED2Vec model, or a hugging face model repository of geniml.region2vec.Region2VecExModel
        """
        if isinstance(query2vec, str):
            self.query2vec = BED2Vec(query2vec)
        else:
            self.query2vec = query2vec

        self.backend = backend

    def query_search(
        self,
        query: Union[str, RegionSet, np.ndarray],
        limit: int,
        with_payload: bool = True,
        with_vectors: bool = True,
        offset: int = 0,
    ) -> List[Dict]:
        """
        :param query: a region set, s path to a BED file in disk, or a region set embedding vector

        for rest of the parameters, check the docstring of QdrantBackend.search() or HNSWBackend.search()
        """
        if isinstance(query, np.ndarray):
            search_vec = query
        else:
            search_vec = self.query2vec.forward(query)

        return self.backend.search(search_vec, limit, with_payload, with_vectors, offset)



import gzip
import os
from io import BytesIO
from pathlib import Path
from typing import Optional

import genomicranges
import pandas as pd

from .const import DEFAULT_CACHE_FOLDER


class BedCacheManager:
    def __init__(self, cache_folder: str):
        self.cache_folder = cache_folder
        self.create_cache_folder()

    def create_cache_folder(self, subfolder_path: Optional[str] = None) -> None:
        """
        Create cache folder if it doesn't exist

        :param subfolder_path: path to the subfolder
        """
        if subfolder_path is None:
            subfolder_path = self.cache_folder

        full_path = os.path.abspath(subfolder_path)
        if not os.path.exists(full_path):
            os.makedirs(full_path)

    @staticmethod
    def process_local_bed_data(file_path: str) -> genomicranges.GenomicRanges:
        """Process a local BED file and return the file content as bytes"""
        with open(file_path, "rb") as local_file:
            file_content = local_file.read()

        gr_bed_local = BedCacheManager.decompress_and_convert_to_genomic_ranges(file_content)

        return gr_bed_local

    @staticmethod
    def decompress_and_convert_to_genomic_ranges(
        content: bytes,
    ) -> genomicranges.GenomicRanges:
        """Decompress a BED file and convert it to a GenomicRanges object"""
        is_gzipped = content[:2] == b"\x1f\x8b"

        if is_gzipped:
            with gzip.GzipFile(fileobj=BytesIO(content), mode="rb") as f:
                df = pd.read_csv(f, sep="\t", header=None, engine="pyarrow")
        else:
            df = pd.read_csv(BytesIO(content), sep="\t", header=None, engine="pyarrow")

        header = [
            "seqnames",
            "starts",
            "ends",
            "name",
            "score",
            "strand",
            "thickStart",
            "thickEnd",
            "itemRgb",
            "blockCount",
        ]
        df.columns = header[: len(df.columns)]
        gr = genomicranges.from_pandas(df)

        return gr


def get_abs_path(path: str = DEFAULT_CACHE_FOLDER, create_folder: bool = True) -> str:
    """
    Get absolute path to the folder and create it if it doesn't exist

    :param path: path to the folder
    :param create_folder: create folder if it doesn't exist

    :return: absolute path to the folder
    """
    absolute_cache_folder = os.path.expandvars(path)
    if create_folder:
        Path(absolute_cache_folder).mkdir(parents=True, exist_ok=True)
    return absolute_cache_folder



import os

MODULE_NAME = "bbclient"

DEFAULT_BEDBASE_API = os.getenv("BEDBASE_API") or "https://api.bedbase.org"
DEFAULT_BEDSET_SUBFOLDER = "bedsets"
DEFAULT_BEDFILE_SUBFOLDER = "bedfiles"
DEFAULT_BEDSET_EXT = ".txt"
DEFAULT_BEDFILE_EXT = ".bed.gz"

BEDFILE_URL_PATTERN = "{bedbase_api}/v1/objects/bed.{bed_id}.bed_file/access/http/bytes"
BEDSET_URL_PATTERN = "{bedbase_api}/v1/bedset/{bedset_id}/bedfiles"
BED_TOKENS_PATTERN = "{bedbase_api}/v1/bed/{bed_id}/tokens/{universe_id}/info"

BBCLIENT_CACHE_ENV = "BBCLIENT_CACHE"

HOME_PATH = os.getenv("HOME")
if not HOME_PATH:
    HOME_PATH = os.path.expanduser("~")
DEFAULT_CACHE_FOLDER = os.getenv(BBCLIENT_CACHE_ENV) or os.path.join(HOME_PATH, ".bbcache/")

DEFAULT_ZARR_FOLDER = "tokens.zarr"
DEFAULT_BUCKET_NAME = "bedbase"
DEFAULT_BUCKET_FOLDER = "bed_files"



import argparse

from geniml.bedshift._version import __version__


class _VersionInHelpParser(argparse.ArgumentParser):
    def format_help(self):
        """Add version information to help text."""
        return (
            "version: {}\n".format(__version__) + super(_VersionInHelpParser, self).format_help()
        )


def build_argparser():
    """
    Builds argument parser.

    :return: argparse.ArgumentParser
    """

    banner = "%(prog)s - randomize BED files"
    additional_description = "\n..."

    parser = _VersionInHelpParser(description=banner, epilog=additional_description)

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {v}".format(v=__version__),
    )

    parser.add_argument("-b", "--bedfile", required=True, help="File path to bed file.")

    parser.add_argument(
        "-l",
        "--chrom-lengths",
        type=str,
        required=False,
        help="TSV text file with one row per chromosomes indicating chromosome sizes",
    )

    parser.add_argument(
        "-g",
        "--genome",
        type=str,
        required=False,
        help="Refgenie genome identifier (used for chrom sizes).",
    )

    parser.add_argument("-d", "--droprate", type=float, default=0.0, help="Droprate parameter")

    parser.add_argument("-a", "--addrate", type=float, default=0.0, help="Addrate parameter")

    parser.add_argument("--addmean", type=float, default=320.0, help="Mean add region length")

    parser.add_argument("--addstdev", type=float, default=30.0, help="Stdev add length")

    parser.add_argument("--addfile", type=str, help="Add regions from a bedfile")

    parser.add_argument(
        "--valid-regions",
        type=str,
        dest="valid_regions",
        help="valid regions in which regions can be randomly added",
    )

    parser.add_argument("-s", "--shiftrate", type=float, default=0.0, help="Shift probability")

    parser.add_argument("--shiftmean", type=float, default=0.0, help="Mean shift")

    parser.add_argument("--shiftstdev", type=float, default=150.0, help="Stdev shift")

    parser.add_argument("--shiftfile", type=str, help="Shift regions from a bedfile")

    parser.add_argument("-c", "--cutrate", type=float, default=0.0, help="Cut probability")

    parser.add_argument(
        "-m",
        "--mergerate",
        type=float,
        default=0.0,
        help="Merge probability. WARNING: will likely create regions that are thousands of base pairs long",
    )

    parser.add_argument("--dropfile", type=str, help="Drop regions from a bedfile")

    parser.add_argument(
        "-o",
        "--outputfile",
        type=str,
        help="output file name (including extension). if not specified, will default to bedshifted_{originalname}.bed",
    )

    parser.add_argument(
        "-r",
        "--repeat",
        type=int,
        default=1,
        help="the number of times to repeat the operation",
    )

    parser.add_argument(
        "-y",
        "--yaml-config",
        dest="yaml_config",
        type=str,
        help="run yaml configuration file",
    )

    parser.add_argument(
        "--seed",
        default=None,
        help="an integer-valued seed for allowing reproducible perturbations",
    )

    return parser


param_msg = """Params:
  chrom.sizes file: {chromsizes}
  shift:
    shift rate: {shiftrate}
    shift mean distance: {shiftmean}
    shift stdev: {shiftstdev}
    shift regions from file: {shiftfile}
  add:
    rate: {addrate}
    add mean length: {addmean}
    add stdev: {addstdev}
    add regions from file: {addfile}
    valid regions: {valid_regions}
  cut rate: {cutrate}
  drop rate: {droprate}
    drop regions from file: {dropfile}
  merge rate: {mergerate}
  outputfile: {outputfile}
  repeat: {repeat}
  yaml_config: {yaml_config}
  seed: {seed}
"""



import logging
import sys
from typing import Dict

import logmuse
from ubiquerg import VersionInHelpParser

from ._version import __version__
from .argparsers import build_distance_argparser as distance_subparser
from .argparsers import build_preprocess_argparser as preprocess_subparser
from .argparsers import build_search_argparser as search_subparser
from .argparsers import build_train_argparser as train_subparser
from .const import *

global _LOGGER
logging.basicConfig(level=logging.INFO)
_LOGGER = logging.getLogger(PKG_NAME)


def build_argparser(parser: VersionInHelpParser = None):
    """
    Parse command-line arguments passed to the pipeline.
    """
    # Argument Parsing
    ###########################################################################
    if parser is None:
        parser = VersionInHelpParser(
            prog=PKG_NAME,
            version=__version__,
            description="%(prog)s - co-embed region sets and labels.",
        )

    # Individual subcommands
    msg_by_cmd = {
        PREPROCESS_CMD: "Preprocess data for training",
        TRAIN_CMD: "Train the StarSpace model",
        DISTANCES_CMD: "Compute distances between regionsets and labels",
        SEARCH_CMD: "Search for regionsets similar to a query",
    }

    sp = parser.add_subparsers(dest="subcommand")
    subparsers: Dict[str, VersionInHelpParser] = {}
    for k, v in msg_by_cmd.items():
        subparsers[k] = sp.add_parser(k, description=v, help=v)

    # build up subparsers for commands
    subparsers[PREPROCESS_CMD] = preprocess_subparser(subparsers[PREPROCESS_CMD])
    subparsers[TRAIN_CMD] = train_subparser(subparsers[TRAIN_CMD])
    subparsers[DISTANCES_CMD] = distance_subparser(subparsers[DISTANCES_CMD])
    subparsers[SEARCH_CMD] = search_subparser(subparsers[SEARCH_CMD])

    return parser


def main(args):
    """MAIN"""
    parser = logmuse.add_logging_options(build_argparser())
    args, _ = parser.parse_known_args()

    _LOGGER = logmuse.logger_via_cli(args, make_root=True)
    _LOGGER.debug(f"versions: bedspace {__version__}")
    _LOGGER.debug(f"Args: {args}")

    if args.command == PREPROCESS_CMD:
        print("Here")
        _LOGGER.info("Preprocessing data for training")

    elif args.command == TRAIN_CMD:
        _LOGGER.info("Training the StarSpace model")

    elif args.command == DISTANCES_CMD:
        _LOGGER.info("Computing distances between regionsets and labels")

    elif args.command == SEARCH_CMD:
        _LOGGER.info("Searching for regionsets similar to a query")

    else:
        _LOGGER.error("Unknown command: {}".format(args.command))
        sys.exit(1)



import os

import numpy as np

from ..likelihood.build_model import ModelLH
from ..utils import read_chromosome_from_bw
from .utils import check_if_uni_flexible, check_if_uni_sorted


class LhModel:
    def __init__(self, model, cove):
        """
        Object with combined information about lh model and coverage
        :param ndarray model: lh model array
        :param ndarray cove: coverage array
        """
        self.model = model
        self.cove = cove

    def __getitem__(self, sliced):
        return self.model[self.cove[sliced[0]], sliced[1]]


def calc_likelihood_hard(
    universe,
    chroms,
    model_lh,
    coverage_folder,
    coverage_prefix,
    name,
    s_index,
    e_index=None,
):
    """
    Calculate likelihood of universe for given type of model
    To be used with binomial model
    :param str  universe: path to universe file
    :param list chroms: list of chromosomes present in model
    :param ModelLH model_lh: likelihood model
    :param coverage_prefix: prefix used in uniwig for creating coverage
    :param coverage_folder: path to a folder with genome coverage by tracks
    :param str name: suffix of model file name, which contains information
     about model type
    :param int s_index: from which position in universe line take assess region
     start position
    :param int e_index: from which position in universe line take assess region
     end position
    :return float: likelihood of universe for given model
    """
    current_chrom = ""
    missing_chrom = ""
    empty_start = 0
    res = 0
    e = 0
    done_chrom = []
    prob_array, cove_array = None, None
    with open(universe) as uni:
        for i in uni:
            e += 1
            i = i.split("\t")
            i[1], i[2] = int(i[1]), int(i[2])
            if i[0] == missing_chrom:
                pass
            else:
                if i[0] != current_chrom:
                    if i[0] in chroms:
                        done_chrom.append(i[0])
                        model_lh.clear_chrom(current_chrom)
                        if e != 1:
                            res += np.sum(prob_array[empty_start:, 0])

                        current_chrom = i[0]
                        model_lh.read_chrom_track(current_chrom, name)
                        prob_model = model_lh[current_chrom]
                        cove_array = read_chromosome_from_bw(
                            os.path.join(coverage_folder, f"{coverage_prefix}_{name}.bw"),
                            current_chrom,
                        )
                        prob_array = LhModel(prob_model[name], cove_array)
                        empty_start = 0
                    else:
                        print(f"Chromosome {i[0]} missing from model")
                        missing_chrom = i[0]
                start = i[s_index]
                if e_index is None:
                    end = i[s_index] + 1
                else:
                    end = i[e_index]
                r1 = np.sum(prob_array[start:end, 1])
                r2 = np.sum(prob_array[empty_start:start, 0])
                res += r1
                res += r2
                empty_start = end
    res += np.sum(prob_array[empty_start:, 0])
    z = set(chroms).difference(set(done_chrom))
    z = list(z)
    for i in z:
        model_lh.read_chrom_track(i, name)
        prob_model = model_lh[i]
        cove_array = read_chromosome_from_bw(
            os.path.join(coverage_folder, f"{coverage_prefix}_{name}.bw"),
            i,
        )
        prob_array = LhModel(prob_model[name], cove_array)
        res += np.sum(prob_array[:, 0])
    return res


def hard_universe_likelihood(model, universe, coverage_folder, coverage_prefix):
    """
    Calculate likelihood of hard universe based on core, start,
    end coverage model
    :param str model: path to file containing model
    :param str universe: path to universe
    :param coverage_prefix: prefix used in uniwig for creating coverage
    :param coverage_folder: path to a folder with genome coverage by tracks
    :return float: likelihood
    """
    check_if_uni_sorted(universe)
    model_lh = ModelLH(model)
    chroms = model_lh.chromosomes_list
    s = calc_likelihood_hard(
        universe, chroms, model_lh, coverage_folder, coverage_prefix, "start", 1
    )
    e = calc_likelihood_hard(
        universe, chroms, model_lh, coverage_folder, coverage_prefix, "end", 2
    )
    c = calc_likelihood_hard(
        universe,
        chroms,
        model_lh,
        coverage_folder,
        coverage_prefix,
        "core",
        1,
        2,
    )
    return sum([s, e, c])


def likelihood_only_core(model_file, universe, coverage_folder, coverage_prefix):
    """
    Calculate likelihood of universe based only on core coverage model
    :param str model_file: path to name containing model
    :param str universe: path to universe
    :param coverage_prefix: prefix used in uniwig for creating coverage
    :param coverage_folder: path to a folder with genome coverage by tracks
    :return float: likelihood
    """
    check_if_uni_sorted(universe)
    model_lh = ModelLH(model_file)
    chroms = model_lh.chromosomes_list
    c = calc_likelihood_hard(
        universe,
        chroms,
        model_lh,
        coverage_folder,
        coverage_prefix,
        "core",
        1,
        2,
    )
    return c


def background_likelihood(start, end, model_start, model_cove, model_end):
    """
    Calculate likelihood of background for given region
    """
    res = np.sum(model_start[start:end, 0])
    res += np.sum(model_cove[start:end, 0])
    res += np.sum(model_end[start:end, 0])
    return res


def weigh_livelihood(start, end, model_process, model_cove, model_out, reverse):
    """
    Calculate weighted likelihood of flexible part of the region
    :param int start: start of the region
    :param int end: end of the region
    :param array model_process: model for analyzed type of flexible region
    :param array model_cove: model for coverage
    :param array model_out: model for flexible region that is not being analyzed
    :param bool reverse: if model_process corespondents to end we have to reverse the weighs
    :return float: likelihood of flexible part of the region
    """
    e_w = 1 / (end - start)  # weights for processed model
    c_w = np.linspace(start=e_w, stop=1, num=(end - start))  # weights for core in processed region
    if reverse:
        c_w = c_w[::-1]
    res = e_w * np.sum(model_process[start:end, 1])
    res += np.sum(c_w * model_cove[start:end, 1])
    res += (1 - e_w) * np.sum(model_process[start:end, 0])
    res += np.sum((1 - c_w) * model_cove[start:end, 0])
    res += np.sum(model_out[start:end, 0])
    return res


def flexible_peak_likelihood(start_s, start_e, end_s, end_e, model_start, model_cove, model_end):
    # core part of the peak
    res = np.sum(model_cove[start_e:end_s, 1])
    res += np.sum(model_start[start_e:end_s, 0])
    res += np.sum(model_end[start_e:end_s, 0])
    # start part of the peak
    res += weigh_livelihood(start_s, start_e, model_start, model_cove, model_end, False)
    # end part of the peak
    res += weigh_livelihood(end_s, end_e, model_end, model_cove, model_start, True)
    return res


def read_coverage(cove_folder, cove_prefix, current_chrom):
    cove_start = read_chromosome_from_bw(
        os.path.join(cove_folder, f"{cove_prefix}_start.bw"),
        current_chrom,
    )
    cove_core = read_chromosome_from_bw(
        os.path.join(cove_folder, f"{cove_prefix}_core.bw"),
        current_chrom,
    )
    cove_end = read_chromosome_from_bw(
        os.path.join(cove_folder, f"{cove_prefix}_end.bw"),
        current_chrom,
    )
    cove = {"start": cove_start, "core": cove_core, "end": cove_end}
    return cove


def likelihood_flexible_universe(
    model_file, universe, cove_folder, cove_prefix, save_peak_input=False
):
    """
    Likelihood of given universe under the model
    :param str model_file: path to file with lh model
    :param str universe: path to universe
    :param cove_folder: path to a folder with genome coverage by tracks
    :param cove_prefix: prefix used in uniwig for creating coverage
    :param bool save_peak_input: whether to save universe with each peak lh
    :return float: lh of the flexible universe
    """
    current_chrom = ""
    missing_chrom = ""
    empty_start = 0
    res = 0
    check_if_uni_sorted(universe)
    check_if_uni_flexible(universe)
    model_lh = ModelLH(model_file)
    chroms = model_lh.chromosomes_list
    done_chroms = []
    output = []
    e = 0  # number of processed chromosomes
    with open(universe) as uni:
        for line in uni:
            i = line.split("\t")
            peak_start_s, peak_end_e = int(i[1]), int(i[2])
            peak_start_e, peak_end_s = int(i[6]), int(i[7])
            if i[0] == missing_chrom:
                pass
            else:
                if i[0] != current_chrom:
                    if i[0] in chroms:
                        model_lh.clear_chrom(current_chrom)
                        if e != 0:
                            # if we read any chromosomes add to result background
                            # likelihood of part of the genome after the last region
                            res += background_likelihood(
                                empty_start,
                                chr_size,
                                prob_start,
                                prob_core,
                                prob_end,
                            )
                        current_chrom = i[0]
                        done_chroms.append(current_chrom)
                        e += 1
                        model_lh.read_chrom(current_chrom)
                        models_current = model_lh[current_chrom]
                        cove_current = read_coverage(cove_folder, cove_prefix, current_chrom)
                        chr_size = len(cove_current["start"])
                        prob_start = LhModel(models_current["start"], cove_current["start"])
                        prob_core = LhModel(models_current["core"], cove_current["core"])
                        prob_end = LhModel(models_current["end"], cove_current["end"])

                    else:
                        print(f"Chromosome {i[0]} missing from model")
                        missing_chrom = i[0]
            res += background_likelihood(
                empty_start,
                peak_start_s,
                prob_start,
                prob_core,
                prob_end,
            )
            peak_likelihood = flexible_peak_likelihood(
                peak_start_s,
                peak_start_e,
                peak_end_s,
                peak_end_e,
                prob_start,
                prob_core,
                prob_end,
            )
            res += peak_likelihood
            if save_peak_input:
                background = background_likelihood(
                    peak_start_s, peak_end_e, prob_start, prob_core, prob_end
                )
                contribution = peak_likelihood - background
                output.append("{}\t{}\n".format(line.strip("\n"), contribution))
            empty_start = peak_end_e

        res += background_likelihood(empty_start, chr_size, prob_start, prob_core, prob_end)
        if save_peak_input:
            print("saving")
            with open(universe + "_peak_likelihood", "w") as f:
                f.writelines(output)
    z = set(chroms).difference(set(done_chroms))
    z = list(z)
    for i in z:
        for name in ["start", "core", "end"]:
            model_lh.read_chrom_track(i, name)
            prob_model = model_lh[i]
            cove_array = read_chromosome_from_bw(
                os.path.join(cove_folder, f"{cove_prefix}_{name}.bw"),
                i,
            )
            prob_array = LhModel(prob_model[name], cove_array)
            res += np.sum(prob_array[:, 0])
    return res



from ubiquerg import VersionInHelpParser

from ._version import __version__
from .const import *


def build_preprocess_argparser(
    parser: VersionInHelpParser,
) -> VersionInHelpParser:
    parser.add_argument(
        "-i",
        "--input",
        dest="input",
        help="Path to input bed files",
    )

    parser.add_argument("-m", "--metadata", dest="metadata", help="Path to metadata file")

    parser.add_argument(
        "-u",
        "--universe",
        dest="universe",
        help="Path to universe file",
    )

    parser.add_argument(
        "-o",
        "--output",
        default="out",
        dest="output",
        help="Path to output folder",
    )

    parser.add_argument(
        "-l",
        "--labels",
        default="cell_type,target",
        dest="labels",
        help="Labels string",
    )


def build_train_argparser(parser: VersionInHelpParser) -> VersionInHelpParser:
    parser.add_argument(
        "-s",
        "--path-to-starsapce",
        dest="path_to_starsapce",
        help="Path to starspace folder (must be prebuilt)",
    )

    parser.add_argument(
        "-i",
        "--input",
        dest="input",
        help="Path to input file generated by preprocess",
    )

    parser.add_argument(
        "-n",
        "--epochs",
        dest="num_epochs",
        help="Number of epochs",
    )

    parser.add_argument(
        "-d",
        "--dim",
        dest="dim",
        help="Dimension of embeddings",
    )

    parser.add_argument(
        "-l",
        "--lr",
        dest="learning_rate",
        help="Learning rate",
    )

    parser.add_argument(
        "-o",
        "--output",
        default="out",
        dest="output",
        help="Path to output folder",
    )


def build_distance_argparser(
    parser: VersionInHelpParser,
) -> VersionInHelpParser:
    parser.add_argument(
        "-i",
        "--input",
        dest="input",
        help="Path to a trained model",
    )

    parser.add_argument(
        "-m",
        "--metadata",
        dest="metadata",
        help="Path to metadata file",
    )

    parser.add_argument(
        "-u",
        "--universe",
        dest="universe",
        help="Path to universe file",
    )

    parser.add_argument(
        "-f",
        "--files",
        dest="files",
        help="Path to files to be embedded (the data)",
    )

    parser.add_argument(
        "-l",
        "--labels",
        dest="labels",
        default="cell_type,target",
        help="Labels string",
    )

    parser.add_argument(
        "-o",
        "--output",
        default="out",
        dest="output",
        help="Path to output folder",
    )

    parser.add_argument(
        "-t",
        "--threshold",
        dest="threshold",
        default=0.5,
        help="Threshold for classification",
    )


def build_search_argparser(parser: VersionInHelpParser) -> VersionInHelpParser:
    parser.add_argument("query", help="Search query")
    parser.add_argument(
        "-t",
        "--type",
        dest="type",
        type=SearchType,
        help="Search type. Can be one of the following: 1) Label2Region (l2r) 2) Region2Label (r2l) 3) Region2Region (r2r)",
    )

    parser.add_argument(
        "-d",
        "--distances",
        dest="distances",
        help="Path to distances file",
    )

    parser.add_argument(
        "-n",
        "--num-results",
        dest="num_results",
        default=10,
        help="Number of results to return",
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
from functools import cmp_to_key

import numpy as np
import pyBigWig

from geniml.utils import natural_chr_sort, timer_func


def ana_region(reg, start_s, starts, ends, track_val):
    """Check how many regions given part of universe contains
    :param ndarray reg: vector with universes states; 0 - background, 1- boundary, 2- core
    :param int start_s: start of the part of the universe
    :param list starts: list of region starts in given part of universe
    :param list ends: list of region ends in given part of universe
    :param ndarray track_val: genome coverage by the collection for given part of universe
    """
    core_s, core_e = [], []
    core_pos = np.argwhere(reg == 2).flatten()
    if len(core_pos) == 0:
        return "empty"
    else:
        core_s.append(core_pos[0] + start_s)
        if core_pos[0] == 0:
            core_s[0] += 1
        for i in range(1, len(core_pos)):
            if core_pos[i] - core_pos[i - 1] >= 50:
                core_e.append(core_pos[i - 1] + start_s)
                min_point = np.argmin(track_val[core_pos[i - 1] : core_pos[i]])
                min_point = min_point + core_pos[i - 1]
                ends.append(int(min_point) + start_s)
                starts.append(ends[-1] + 1)
                core_s.append(core_pos[i] + start_s)
                if core_s[-1] == starts[-1]:
                    core_s[-1] += 1
        core_e.append(core_pos[-1] + start_s)
        if core_pos[-1] == len(reg) - 1:
            core_e[-1] -= 1
    return core_s, core_e, starts, ends


def save_regions(inter_pos, chrom, bedname, track):
    """Save regions from universes to file
    :param ndarray inter_pos: vector with universes states; 0 - background, 1- boundary, 2- core
    :param str chrom: chromosome to analyse
    :param str bedname: output file
    :param ndarray track: vector with coverage values
    """
    ind = np.argwhere(inter_pos != 0)
    ind = ind.flatten()
    start_s = ind[0]
    to_file = []
    line = chrom + "\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n"
    for i in range(1, len(ind)):
        if ind[i] - ind[i - 1] != 1:
            end_e = ind[i - 1]
            region = inter_pos[start_s : end_e + 1]
            res = ana_region(region, start_s, [start_s], [], track[start_s : end_e + 1])
            if res != "empty":
                save_start_e, save_end_s, save_start_s, save_end_e = res
                save_end_e = save_end_e + [end_e]
                for a, b, c, d in zip(save_start_e, save_end_s, save_start_s, save_end_e):
                    if a != b:
                        val = 0
                        li = line.format(
                            int(c),
                            int(d) + 1,
                            "universe",
                            val,
                            ".",
                            int(a),
                            int(b),
                            "0,0,255",
                        )
                        to_file.append(li)
            start_s = ind[i]
    end_e = ind[-1]
    region = inter_pos[start_s : end_e + 1]
    res = ana_region(region, start_s, [start_s], [], track[start_s : end_e + 1])
    if res != "empty":
        save_start_e, save_end_s, save_start_s, save_end_e = res
        save_end_e = save_end_e + [end_e]
        for a, b, c, d in zip(save_start_e, save_end_s, save_start_s, save_end_e):
            val = 0
            li = line.format(
                int(c),
                int(d) + 1,
                "universe",
                val,
                ".",
                int(a),
                int(b),
                "0,0,255",
            )
            to_file.append(li)
    with open(bedname, "a") as f:
        f.writelines(to_file)


def get_uni(file, chrom, bedname):
    """Build cut-off coverage flexible universes from coverage track
    :param str file: coverage file
    :param str chrom: chromosome to analyse
    :param str bedname: output file
    """
    file = pyBigWig.open(file)
    if pyBigWig.numpy:
        track = file.values(chrom, 0, file.chroms(chrom), numpy=True)
    else:
        track = file.values(chrom, 0, file.chroms(chrom))
        track = np.array(track)
    track[np.isnan(track)] = 0
    track = track.astype(np.uint16)
    cutoff = np.sum(track) / len(track)
    track_non_zero_sort = np.sort(track[track != 0])

    cutoff = max([1, np.round(cutoff)])
    pos = np.where(track_non_zero_sort == cutoff)[0]
    if len(pos) == 0:
        uniq_val = np.unique(track_non_zero_sort)
        dist = np.absolute(uniq_val - cutoff)
        cutoff = uniq_val[dist.argmin()]
        pos = np.where(track_non_zero_sort == cutoff)[0]
    f, l = pos[0] / len(track_non_zero_sort), pos[-1] / len(track_non_zero_sort)
    q_cutoff = np.mean([f, l])
    lower = np.quantile(track_non_zero_sort, max(0, q_cutoff - 0.2))
    upper = np.quantile(track_non_zero_sort, min(1, q_cutoff + 0.2))
    inter_pos = np.zeros(len(track), dtype=np.uint8)
    inter_pos[track >= lower] = 1
    inter_pos[track > upper] = 2
    save_regions(inter_pos, chrom, bedname, track)


def ccf_universe(cove, file_out, cove_prefix="all"):
    """
    Create cut-off flexible universe based on coverage track
    :param str cove: path to coverage folder
    :param str file_out: output file
    :param str cove_prefix: prefixed used for creating signal tracks
    """
    if os.path.isfile(file_out):
        raise Exception(f"File : {file_out} exists")
    file = os.path.join(cove, f"{cove_prefix}_core.bw")
    bw = pyBigWig.open(file)
    chroms = bw.chroms()
    bw.close()
    chroms_key = list(chroms.keys())
    chroms_key = sorted(chroms_key, key=cmp_to_key(natural_chr_sort))
    chroms = {i: chroms[i] for i in chroms_key}
    for chrom in chroms:
        if chroms[chrom] > 0:
            get_uni(file, chrom, file_out)



import os
import time
from typing import List

import numpy as np
import scanpy as sc
from rich.progress import track

from ..io import Region


class Timer:
    """Records the running time.

    Uses Timer.s() or Timer() to record the start time. Then, calls Timer.t() to get the
    elapsed time in seconds.
    """

    def __init__(self):
        """Initializes a Timer object and starts the timer."""
        self.v = time.time()

    def s(self):
        """Restarts the timer."""
        self.v = time.time()

    def t(self):
        """Gives the elapsed time.

        Returns:
            float: The elapsed time in seconds.
        """
        return time.time() - self.v


def time_str(t: float) -> str:
    """Converts time in float to a readable format.

    Converts time in float to hours, minutes, or seconds based on the value of
    t.

    Args:
        t (float): Time in seconds.

    Returns:
        str: Time in readable time.
    """
    if t >= 3600:
        return f"{t / 3600:.2f}h"
    if t >= 60:
        return f"{t / 60:.2f}m"
    return f"{t:.2f}s"


def anndata_to_regionsets(adata: sc.AnnData) -> List[List[Region]]:
    """
    Converts an AnnData object to a list of lists of regions. This
    is done by taking each cell and creating a list of all regions
    that have a value greater than 0.

    This function is already pretty optimized. To speed this up
    further we'd have to parallelize it or reach for a lower-level
    language.

    *Note: this method requires that the sc.AnnData object have
    chr, start, and end in `.var` attributes*
    """
    if not isinstance(adata, sc.AnnData):
        raise ValueError("The input must be a scanpy AnnData object.")

    if not all(
        [
            "chr" in adata.var.columns,
            "start" in adata.var.columns,
            "end" in adata.var.columns,
        ]
    ):
        raise ValueError(
            "The AnnData object must have chr, start, and end in the `.var` attribute."
        )

    # Extract the arrays for chr, start, and end
    chr_values = adata.var["chr"].values
    start_values = adata.var["start"].values
    end_values = adata.var["end"].values

    # Perform the comparison using numpy operations
    positive_values = adata.X > 0

    if not isinstance(positive_values, np.ndarray):
        positive_values = positive_values.toarray()

    regions = []
    for i in track(range(adata.shape[0]), total=adata.shape[0], description="Tokenizing"):
        regions.append(
            [
                Region(chr_values[j], start_values[j], end_values[j])
                for j in np.where(positive_values[i])[0]
            ]
        )
    return regions



def build_subparser_hmm(parser):
    parser = build_subparser(parser)

    parser.add_argument(
        "--not-normalize",
        help="if not to normalize coverage signal before using HMM",
        action="store_false",
    )
    parser.add_argument(
        "--save-max-cove",
        help="if present saves maximum coverage for each peak",
        action="store_true",
    )

    return parser


def build_subparser_ml(parser):
    parser = build_subparser(parser)
    parser.add_argument("--model-file", help="path to lh model file", required=True, type=str)
    return parser


def build_subparser_cc(parser):
    parser = build_subparser(parser)
    parser.add_argument(
        "--merge",
        help="distance between output peaks that should be merged into one in output universe",
        default=0,
        type=int,
    )
    parser.add_argument(
        "--filter-size",
        help="minimal size of the region in the universe",
        default=0,
        type=int,
    )
    parser.add_argument("--cutoff", help="cutoff value used for making universe", type=int)

    return parser


def build_subparser(parser):
    """
    Parse command-line arguments passed to the pipeline.
    """
    parser.add_argument(
        "--output-file",
        help="path to output, universe file",
        required=True,
        type=str,
    )
    parser.add_argument(
        "--coverage-folder",
        help="path to core coverage folder",
        required=True,
        type=str,
    )
    parser.add_argument(
        "--coverage-prefix",
        help="prefixed used for making coverage files",
        default="all",
        type=str,
    )

    return parser


def build_mode_parser(parser):
    sp = parser.add_subparsers(dest="subcommand")
    msg_by_cmd = {
        "cc": "Making coverage cut-off universe",
        "ccf": "Making coverage cut-off flexible universe",
        "ml": "Making ML universe",
        "hmm": "Making HMM universe",
    }
    subparsers = {}
    for k, v in msg_by_cmd.items():
        subparsers[k] = sp.add_parser(k, description=v, help=v)
    subparsers["cc"] = build_subparser_cc(subparsers["cc"])
    subparsers["ccf"] = build_subparser(subparsers["ccf"])
    subparsers["ml"] = build_subparser_ml(subparsers["ml"])
    subparsers["hmm"] = build_subparser_hmm(subparsers["hmm"])
    return parser



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import tempfile
from multiprocessing import Pool

import numpy as np
import pandas as pd

from ..utils import natural_chr_sort
from .utils import check_if_uni_flexible, check_if_uni_sorted, prep_data, process_db_line


def flexible_distance_between_two_regions(region, query):
    """Calculate distance between region and flexible region from flexible universe
    :param [int, int] region: region from flexible universe
    :param int query: analyzed region
    :return int: distance
    """
    if region[0] <= query <= region[1]:
        return 0
    else:
        return min(abs(region[0] - query), abs(region[1] - query))


def distance_between_two_regions(region, query):
    """Calculate distance between region in database and region from the query
    :param [int] region: region from hard universe
    :param int query: analysed region
    :return int: distance
    """
    return abs(region[0] - query)


def distance_to_closest_region(
    db, db_queue, i, current_chrom, unused_db, pos_index, flexible, uni_to_file
):
    """
    Calculate distance from given peak to the closest region in database
    :param file db: database file
    :param list db_queue: queue of three last positions in database
    :param i: analyzed position from the query
    :param str current_chrom: current analyzed chromosome from query
    :param list unused_db: list of positions from universe that were not compared to query
    :param list pos_index: which indexes from universe region use to calculate distance
    :param bool flexible: whether the universe if flexible
    :param bool uni_to_file: whether calculate distance from universe to file
    :return int: peak distance to universe
    """
    if flexible:
        if uni_to_file:
            dist_to_db_que = [flexible_distance_between_two_regions(i, j[0]) for j in db_queue]
        else:
            dist_to_db_que = [flexible_distance_between_two_regions(j, i[0]) for j in db_queue]
    else:
        dist_to_db_que = [distance_between_two_regions(j, i[0]) for j in db_queue]
    min_pos = np.argmin(dist_to_db_que)
    while min_pos == 2:
        d = db.readline().strip("\n")
        if d == "":
            return dist_to_db_que[min_pos]
        pos, pos_chrom = process_db_line(d, pos_index)
        if pos_chrom != current_chrom:
            unused_db.append([pos, pos_chrom])
            return dist_to_db_que[min_pos]
        db_queue[:-1] = db_queue[1:]
        db_queue[-1] = pos
        if flexible:
            if uni_to_file:
                dist_to_db_que = [flexible_distance_between_two_regions(i, j[0]) for j in db_queue]
            else:
                dist_to_db_que = [flexible_distance_between_two_regions(j, i[0]) for j in db_queue]
        else:
            dist_to_db_que = [distance_between_two_regions(j, i[0]) for j in db_queue]
        min_pos = np.argmin(dist_to_db_que)
    return dist_to_db_que[min_pos]


def read_in_new_universe_regions(
    db,
    q_chrom,
    current_chrom,
    unused_db,
    db_queue,
    waiting,
    pos_index,
):
    """
    Read in new universe regions closest to the peak
    :param file db: universe file
    :param str q_chrom: new peak's chromosome
    :param str current_chrom: chromosome that was analyzed so far
    :param list unused_db: list of positions from universe that were not compared to query
    :param list db_queue: que of three last positions in universe
    :param bool waiting: whether iterating through file, without calculating
     distance,  if present chromosome not present in universe
    :param list pos_index: which indexes from universe region use to calculate distance
    :return bool, str: if iterating through chromosome not present in universe; current chromosome in query
    """
    if q_chrom != current_chrom:
        # change chromosome
        db_queue.clear()
        # clean up the que
        if len(unused_db) == 0:
            d = db.readline().strip("\n")
            if d == "":
                waiting = True
                return waiting, current_chrom
            d_start, d_start_chrom = process_db_line(d, pos_index)
            while current_chrom == d_start_chrom:
                # finish reading old chromosome in DB file
                d = db.readline().strip("\n")
                if d == "":
                    break
                d_start, d_start_chrom = process_db_line(d, pos_index)
            unused_db.append([d_start, d_start_chrom])
        current_chrom = q_chrom
        if current_chrom == unused_db[-1][1]:
            waiting = False
            db_queue.append(unused_db[-1][0])
            unused_db.clear()
        elif natural_chr_sort(unused_db[-1][1], current_chrom) == 1:
            # chrom present in file not in DB
            waiting = True
            return waiting, current_chrom
        while len(db_queue) < 3:
            d = db.readline().strip("\n")
            if d == "":
                break
            d_start, d_start_chrom = process_db_line(d, pos_index)
            if d_start_chrom == current_chrom:
                db_queue.append(d_start)
            elif natural_chr_sort(d_start_chrom, current_chrom) == 1:
                unused_db.append([d_start, d_start_chrom])
                waiting = True
                return waiting, current_chrom
    if len(db_queue) == 0:
        waiting = True
    return waiting, current_chrom


def calc_distance_between_two_files(
    universe,
    q_folder,
    q_file,
    flexible,
    save_each,
    folder_out,
    pref,
    uni_to_file=False,
):
    """
    Maine function for calculating distance between regions in file query to regions in database
    :param str universe: path to universe
    :param str q_folder: path to folder containing query files
    :param str q_file: query file
    :param boolean flexible: whether the universe if flexible
    :param bool save_each: whether to save calculated distances for each file
    :param str folder_out: output folder
    :param str pref: prefix used as the name of the folder
     containing calculated distance for each file
    :param uni_to_file: whether to calculate distance from universe to file
    :return str, int, int: file name; median od distance of starts to
     starts in universe; median od distance of ends to ends in universe
    """
    query = tempfile.NamedTemporaryFile()
    prep_data(q_folder, q_file, query)
    if uni_to_file:
        db_start_name = query.name
        db_end_name = query.name
        q = open(universe)
    else:
        db_start_name = universe
        db_end_name = universe
        q = query
    with open(db_start_name) as db_start, open(db_end_name) as db_end:
        db_queue_start = []
        current_chrom_start = "chr0"
        dist_start = []
        unused_db_start = []
        waiting_start = False
        db_queue_end = []
        current_chrom_end = "chr0"
        dist_end = []
        unused_db_end = []
        waiting_end = False
        start_index_q, start_index_db = [1], [1]
        end_index_q, end_index_db = [2], [2]
        if flexible and uni_to_file:
            start_index_q = [1, 6]
            end_index_q = [7, 2]
        if flexible and not uni_to_file:
            start_index_db = [1, 6]
            end_index_db = [7, 2]
        for i in q:
            if not uni_to_file:
                i = i.decode("utf-8")
            i = i.split("\t")
            start = [int(i[ind]) for ind in start_index_q]
            end = [int(i[ind]) for ind in end_index_q]
            q_chrom = i[0]
            result_start = read_in_new_universe_regions(
                db_start,
                q_chrom,
                current_chrom_start,
                unused_db_start,
                db_queue_start,
                waiting_start,
                start_index_db,
            )
            (waiting_start, current_chrom_start) = result_start
            if not waiting_start:
                result = distance_to_closest_region(
                    db_start,
                    db_queue_start,
                    start,
                    current_chrom_start,
                    unused_db_start,
                    start_index_db,
                    flexible,
                    uni_to_file,
                )
                dist_start.append(result)
            result_end = read_in_new_universe_regions(
                db_end,
                q_chrom,
                current_chrom_end,
                unused_db_end,
                db_queue_end,
                waiting_end,
                end_index_db,
            )
            (waiting_end, current_chrom_end) = result_end
            if not waiting_end:
                res = distance_to_closest_region(
                    db_end,
                    db_queue_end,
                    end,
                    current_chrom_end,
                    unused_db_end,
                    end_index_db,
                    flexible,
                    uni_to_file,
                )
                dist_end.append(res)
    query.close()
    if save_each:
        with open(os.path.join(folder_out, pref, q_file), "w") as f:
            for i, j in zip(dist_start, dist_end):
                f.write(f"{i}\t{j}\n")
    if not dist_start:
        print(f"File {q_file} doesn't contain any chromosomes present in universe")
        return q_file, None
    dist = dist_start + dist_end
    return q_file, np.median(dist)


def run_distance(
    folder,
    file_list,
    universe,
    no_workers,
    flexible=False,
    folder_out=None,
    pref=None,
    save_each=False,
    uni_to_file=False,
):
    """
    For group of files calculate distance to the nearest region in universe
    :param str folder: path to folder containing query files
    :param str file_list: path to file containing list of query files
    :param str universe: path to universe file
    :param int no_workers: number of parallel processes
    :param bool flexible: whether the universe if flexible
    :param str folder_out: output folder
    :param str pref: prefix used for saving
    :param bool save_each: whether to save calculated distances for each file
    :param uni_to_file: whether to calculate distance from universe to file
    :return float; float: mean of median distances from starts in query to the nearest starts in universe;
    mean of median distances from ends in query to the nearest ends in universe
    """
    check_if_uni_sorted(universe)
    if flexible:
        check_if_uni_flexible(universe)
    with open(file_list) as f:
        files = f.read().split("\n")[:-1]
    res = []
    if save_each:
        os.makedirs(os.path.join(folder_out, pref), exist_ok=True)
    if no_workers <= 1:
        for i in files:
            r = calc_distance_between_two_files(
                universe,
                folder,
                i,
                flexible,
                save_each,
                folder_out,
                pref,
                uni_to_file,
            )
            res.append(r)
    else:
        with Pool(no_workers) as p:
            args = [
                (
                    universe,
                    folder,
                    f,
                    flexible,
                    save_each,
                    folder_out,
                    pref,
                    uni_to_file,
                )
                for f in files
            ]
            res = p.starmap(calc_distance_between_two_files, args)
    return pd.DataFrame(res)



import logging
import math
import os
from typing import List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np

try:
    import torch
    from torch.nn import CosineEmbeddingLoss, CosineSimilarity, Linear, MSELoss, ReLU, Sequential
except ImportError:
    raise ImportError(
        "Please install Machine Learning dependencies by running 'pip install geniml[ml]'"
    )
from huggingface_hub import hf_hub_download
from yaml import safe_dump, safe_load

from .const import (
    CONFIG_FILE_NAME,
    DEFAULT_BATCH_SIZE,
    DEFAULT_HUGGINGFACE_MODEL_NAME,
    DEFAULT_LEARNING_RATE,
    DEFAULT_LOSS_NAME,
    DEFAULT_MARGIN,
    DEFAULT_MUST_TRAINED,
    DEFAULT_NUM_EPOCHS,
    DEFAULT_NUM_UNITS,
    DEFAULT_OPTIMIZER_NAME,
    DEFAULT_PATIENCE,
    DEFAULT_PLOT_FILE_NAME,
    DEFAULT_PLOT_TITLE,
    MODULE_NAME,
    TORCH_MODEL_FILE_NAME_PATTERN,
)
from .utils import arrays_to_torch_dataloader, dtype_check

_LOGGER = logging.getLogger(MODULE_NAME)


class Vec2Vec(Sequential):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_units: Union[int, List[int]],
    ):
        if not isinstance(num_units, list):
            num_units = [num_units]
        num_extra_hidden_layers = len(num_units) - 1
        # input and first hidden layer
        current_layer_units_num = num_units[0]
        layers_list = [
            Linear(in_features=input_dim, out_features=current_layer_units_num),
            ReLU(),
        ]
        previous_layer_units_num = current_layer_units_num

        # extra hidden layer
        for i in range(num_extra_hidden_layers):
            current_layer_units_num = num_units[i + 1]
            layers_list.append(
                Linear(
                    in_features=previous_layer_units_num,
                    out_features=current_layer_units_num,
                )
            )
            layers_list.append(ReLU())
            previous_layer_units_num = current_layer_units_num

        # output layer
        layers_list.append(Linear(in_features=previous_layer_units_num, out_features=output_dim))

        super().__init__(*layers_list)


class Vec2VecFNN:
    def __init__(self, model_path: Union[str, None] = None):
        """
        Initialize Vec2VecFNNtorch.

        :param model_path: path to the pretrained model on huggingface.
        """
        # initialize the feedforward neural network model, which is a torch.nn.Sequential
        # self.model =
        self.model = None
        # whether the model is trained
        self.trained = False
        # optimizer
        self.optimizer = None
        # loss function
        self.loss_fn = None
        # model configure
        self.config = {}
        # training history
        self.most_recent_train = {}

        if model_path is not None:
            # load from hugging face
            self._init_from_huggingface(model_path)
            self.trained = True

    def _init_from_huggingface(
        self,
        model_path: str,
        model_file_name: str = DEFAULT_HUGGINGFACE_MODEL_NAME,
        config_file_name: str = CONFIG_FILE_NAME,
        **kwargs,
    ):
        """
        Initialize the model from a huggingface model. This uses the model path
        to download the necessary files and then "build itself up" from those.

        :param model_path: path to the pre-trained model on huggingface.
        :param model_file_name: name of the model file.
        :param config_file_name: name of the config file
        """
        model_file_path = hf_hub_download(model_path, model_file_name, **kwargs)
        config_path = hf_hub_download(model_path, config_file_name, **kwargs)

        self.load_from_disk(model_file_path, config_path)

    def load_from_disk(self, model_path: str, config_path: str):
        """
        Load model from local files

        :param model_path: path of saved model file (usually in format of .pt)
        :param config_path: path of saved config file (in format of yaml)
        """
        # get the model config (layer structure)
        with open(config_path, "r") as f:
            config = safe_load(f)

        self.config = config
        # reinitiate the self.model
        self.model = Vec2Vec(
            config["input_dim"],
            config["output_dim"],
            config["num_units"],
        )

        # load the Sequential model from saved files
        self.model.load_state_dict(torch.load(model_path))

    def export(
        self,
        path: str,
        checkpoint_file: str,
        config_file: str = CONFIG_FILE_NAME,
        must_trained: bool = DEFAULT_MUST_TRAINED,
    ):
        """
        Save model weights and config

        :param path: path to export the model to
        :param checkpoint_file: name of model checkpoint file
        :param config_file: name of model config file
        :param must_trained: whether the model needs training to be exported
        """
        # whether the model must be finished training to export
        if must_trained and not self.trained:
            raise RuntimeError("Cannot export an untrained model.")

        if not os.path.exists(path):
            os.makedirs(path)

        # export the model weights
        torch.save(self.model.state_dict(), os.path.join(path, checkpoint_file))

        # export the model config
        with open(os.path.join(path, config_file), "w") as f:
            safe_dump(self.config, f)

    def embedding_to_embedding(self, input_vecs: np.ndarray) -> np.ndarray:
        """
        Predict the region set embedding from embedding of natural language strings

        :param input_vecs: input embedding vectors
        :return: the output of the neural network model
        """
        # pytorch tensor's default dtype is float 32
        return self.model(torch.from_numpy(dtype_check(input_vecs))).detach().numpy()

    def compile(
        self,
        optimizer: str,
        loss: str,
        learning_rate: float,
        margin: Union[float, None] = DEFAULT_MARGIN,
    ):
        """
        Configure the model for training. This includes setting the optimizer and loss function.

        :param optimizer: the name of optimizer
        :param loss: the name of loss function
        :param learning_rate: the learning rate of model backpropagation
        :param margin: should be a number from −1−1 to 1, 0 to 0.5 is suggested, only for CosineEmbeddingLoss
        """

        # set optimizer
        if optimizer == "Adam":
            self.optimizer = torch.optim.Adam(self.model.parameters(), learning_rate)

        elif optimizer == "SGD":
            self.optimizer = torch.optim.SGD(self.model.parameters(), learning_rate)

        else:
            raise ValueError("Please give a valid name of optimizer")

        # set loss function
        if loss == "cosine_embedding_loss":
            self.loss_fn = CosineEmbeddingLoss(margin=margin)
        elif loss == "cosine_similarity":
            self.loss_fn = CosineSimilarity()
        elif loss == "mean_squared_error":
            self.loss_fn = MSELoss()
        else:
            raise ValueError("Please give a valid name of loss function")

        # add information to model config
        self.config["optimizer"] = optimizer
        self.config["loss"] = loss

    def train(
        self,
        training_X: np.ndarray,
        training_Y: np.ndarray,
        validating_data: Union[Tuple[np.ndarray, np.ndarray], None] = None,
        save_best: bool = False,
        folder_path: Union[str, None] = None,
        best_model_file_name: Union[str, None] = None,
        early_stop: bool = False,
        patience: float = DEFAULT_PATIENCE,
        opt_name: str = DEFAULT_OPTIMIZER_NAME,
        loss_func: str = DEFAULT_LOSS_NAME,
        num_epochs: int = DEFAULT_NUM_EPOCHS,
        batch_size: int = DEFAULT_BATCH_SIZE,
        learning_rate: float = DEFAULT_LEARNING_RATE,
        training_target: Union[np.ndarray, None] = None,
        validating_target: Union[np.ndarray, None] = None,
        **kwargs,
    ):
        """
        Based on https://pytorch.org/tutorials/beginner/introyt/trainingyt.html
        Fit the feedforward neural network

        :param training_X: embedding vectors of metadata, np.ndarray with shape of (n, <dim>)
        :param training_Y: embedding vectors of region set, np.ndarray with shape of (n, <dim>)
        :param validating_data: validating data, which contains validating X and validating Y
        :param save_best: whether the best performance model is saved after each epoch (based on validation loss)
        :param folder_path: the path to the folder to save the model and config
        :param best_model_file_name: the name of the file of saved best model
        :param early_stop: whether the training should be stopped early to prevent overfitting
        :param patience: the percentage of epoches to stop training if no validation loss improvement
        :param opt_name: name of optimizer
        :param loss_func: name of loss function
        :param num_epochs: number of training epoches
        :param batch_size: size of batch for training
        :param learning_rate: learning rate of optimizer
        :param training_target:
        :param validating_target:
        :param kwargs: see units and layers in reinit_model()
        """
        # if current model is empty, add layers
        if self.model is None:
            # dimensions of input and output
            input_dim = training_X.shape[1]
            output_dim = training_Y.shape[1]

            self.config["input_dim"] = input_dim
            self.config["output_dim"] = output_dim
            self.config["num_units"] = kwargs.get("num_units") or DEFAULT_NUM_UNITS
            self.model = Vec2Vec(
                input_dim=input_dim,
                output_dim=output_dim,
                num_units=self.config["num_units"],
            )

        if training_target is None:
            training_target = np.repeat(1, training_X.shape[0])
        # raise the error if validating data is needed but not provided
        if validating_data is not None:
            validating_X, validating_Y = validating_data
            if validating_target is None:
                validating_target = np.repeat(1, validating_X.shape[0])
            validating_data = arrays_to_torch_dataloader(
                validating_X,
                validating_Y,
                validating_target,
                batch_size=batch_size,
                shuffle=False,
            )

            self.most_recent_train["val_loss"] = []
        elif save_best or early_stop:
            raise ValueError("Validating data is not provided")
        if save_best and folder_path is None:
            raise ValueError(
                "ValueError: Path to folder where the best performance model will be saved is required"
            )

        # compile the model
        self.compile(optimizer=opt_name, loss=loss_func, learning_rate=learning_rate)

        # convert training data from np.ndarray to DataLoader
        training_data = arrays_to_torch_dataloader(
            training_X, training_Y, training_target, batch_size
        )

        best_val_loss = 1_000_000.0
        patience_count = 0
        self.most_recent_train["loss"] = []

        for epoch in range(num_epochs):
            # gradient tracking is on
            self.model.train(True)
            avg_loss = self.train_one_epoch(training_data)
            self.most_recent_train["loss"].append(avg_loss)
            # set model to evaluation mode, disabling dropout and using population
            # statistics for batch normalization.
            self.model.eval()

            if validating_data is not None:
                running_val_loss = 0.0
                # disable gradient computation
                with torch.no_grad():
                    for i, (val_x, val_y, val_target) in enumerate(validating_data):
                        val_output = self.model(val_x)
                        val_loss = self.calc_loss(val_output, val_y, val_target)
                        running_val_loss += val_loss

                avg_val_loss = running_val_loss / (i + 1)
                self.most_recent_train["val_loss"].append(avg_val_loss)
                # logging training and validating loss
                _LOGGER.info(f"EPOCH {epoch + 1}: loss: -{avg_loss} - val_loss: -{avg_val_loss}")

                # save the best-performing model
                if avg_val_loss < best_val_loss:
                    # reset the patience count
                    patience_count = 0
                    best_val_loss = avg_val_loss
                    if save_best:
                        self.export(
                            folder_path,
                            best_model_file_name
                            or TORCH_MODEL_FILE_NAME_PATTERN.format(
                                callback="best", checkpoint=str(epoch)
                            ),
                            must_trained=False,
                        )

                # early stop to prevent overfitting
                if early_stop:
                    if avg_val_loss > avg_loss:
                        patience_count += 1
                    if patience_count > int(math.ceil(patience * num_epochs)):
                        break

            else:
                _LOGGER.info(f"EPOCH {epoch + 1}: loss: -{avg_loss}")
        self.trained = True

    def train_one_epoch(self, training_data) -> torch.Tensor:
        """
        Based on https://pytorch.org/tutorials/beginner/introyt/trainingyt.html

        One epoch's training loop

        :return: the average training loss of one epoch
        """
        epoch_loss = 0.0

        # train on each batch
        for i, (x, y, target) in enumerate(training_data):
            # zero gradients for every batch
            self.optimizer.zero_grad()

            # batch prediction
            outputs = self.model(x)

            # compute loss and gradients
            batch_loss = self.calc_loss(outputs, y, target)
            batch_loss.backward()

            # adjust learning weights
            self.optimizer.step()

            # gather loss and report
            epoch_loss += batch_loss.item()
        return epoch_loss / (i + 1)

    def calc_loss(
        self, outputs: torch.Tensor, y: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        """
        Calculating loss when different loss function is given

        :param outputs: the output of model
        :param y: the correct label
        :param target:
        :return: the loss

        """

        if not self.config["loss"]:
            raise ValueError("Please compile the model first")

        # when all targets are 1
        # loss = 1 - cos(output, y)
        # https://pytorch.org/docs/stable/generated/torch.nn.CosineEmbeddingLoss.html
        elif self.config["loss"] == "cosine_embedding_loss":
            return self.loss_fn(outputs, y, target)
        else:
            return self.loss_fn(outputs, y)

    def plot_training_hist(
        self,
        save_path: Union[str, None] = None,
        plot_file_name: Union[str, None] = DEFAULT_PLOT_FILE_NAME,
        title: Union[str, None] = DEFAULT_PLOT_TITLE,
    ) -> None:
        """
        Plot the training & validating loss of the most recent training

        :param save_path: the path of folder where image will be saved
        :param plot_file_name: the file name of the png file
        :param title: the title in the image
        :return: None
        """

        epoch_range = range(1, len(self.most_recent_train["loss"]) + 1)
        train_loss = self.most_recent_train["loss"]
        plt.figure()
        plt.plot(epoch_range, train_loss, "r", label="Training loss")
        try:
            valid_loss = self.most_recent_train["val_loss"]
            plt.plot(epoch_range, valid_loss, "b", label="Validation loss")
        except:
            pass
        plt.title(title)
        plt.legend()
        if save_path:
            plt.savefig(os.path.join(save_path, plot_file_name))
        else:
            plt.show()
        plt.close()

    def __repr__(self):
        return f"Vec2Vec(input_dimension={self.config['input_dim']}, output_dimension={self.config['output_dim']}, trained={self.trained})"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import importlib.util
import os
from functools import cmp_to_key

import numpy as np

from geniml.likelihood.build_model import ModelLH

from ..utils import natural_chr_sort, read_chromosome_from_bw, timer_func
from .utils import find_full, predictions_to_bed

package_name = "numba"

if importlib.util.find_spec(package_name) is None:

    def process_part(
        cove,
        model_start=np.array([[]]),
        model_core=np.array([[]]),
        model_end=np.array([[]]),
    ):
        """
        Finding ML path through matrix using dynamic programing without numba
        :param ndarray cove: coverage tracks
        :param ndarray model_start: lh model for starts
        :param ndarray model_core: lh model for core
        :param ndarray model_end: lh model for ends
        :return ndarray: ML path through matrix
        """
        mat = np.zeros((len(cove), 4))
        (N, M) = mat.shape
        start_b = model_start[cove[:, 0], 0]
        core_b = model_core[cove[:, 1], 0]
        end_b = model_end[cove[:, 2], 0]
        start = model_start[cove[:, 0], 1]
        core = model_core[cove[:, 1], 1]
        end = model_end[cove[:, 2], 1]
        mat[:, 0] = start + core_b + end_b
        mat[:, 1] = start_b + core + end_b
        mat[:, 2] = start_b + core_b + end
        mat[:, 3] = start_b + core_b + end_b

        for i in range(1, N):
            for j in range(M):
                mat[i, j] += max(mat[i - 1, j], mat[i - 1, j - 1])
        path = np.zeros(len(mat), dtype=np.int8)
        path[-1] = np.argmax(mat[-1])
        for i in range(len(mat) - 2, -1, -1):
            prev_index = path[i + 1]
            new_index = prev_index - (mat[i, prev_index - 1] > mat[i, prev_index])
            if new_index == -1:
                new_index = 3
            path[i] = new_index
        return path

else:
    from numba import njit

    @njit
    def process_part(
        cove,
        model_start=np.array([[]]),
        model_core=np.array([[]]),
        model_end=np.array([[]]),
    ):
        """
        Finding ML path through matrix using dynamic programing with numba
        :param ndarray cove: coverage tracks
        :param ndarray model_start: lh model for starts
        :param ndarray model_core: lh model for core
        :param ndarray model_end: lh model for ends
        :return ndarray: ML path through matrix
        """
        mat = np.zeros((len(cove), 4))
        (N, M) = mat.shape
        for i in range(N):
            start_b = model_start[cove[i, 0], 0]
            core_b = model_core[cove[i, 1], 0]
            end_b = model_end[cove[i, 2], 0]
            start = model_start[cove[i, 0], 1]
            core = model_core[cove[i, 1], 1]
            end = model_end[cove[i, 2], 1]
            mat[i, 0] = start + core_b + end_b
            mat[i, 1] = start_b + core + end_b
            mat[i, 2] = start_b + core_b + end
            mat[i, 3] = start_b + core_b + end_b
        for i in range(1, N):
            for j in range(M):
                mat[i, j] += max(mat[i - 1, j], mat[i - 1, j - 1])
        path = np.zeros(len(mat), dtype=np.int8)
        path[-1] = np.argmax(mat[-1])
        for i in range(len(mat) - 2, -1, -1):
            prev_index = path[i + 1]
            new_index = prev_index - (mat[i, prev_index - 1] > mat[i, prev_index])
            if new_index == -1:
                new_index = 3
            path[i] = new_index
        return path


def make_ml_flexible_universe(model_lh, cove_folder, cove_prefix, chrom, file_out):
    """
    Make ML flexible universe per chromosome
    :param ModelLH model_lh: lh model
    :param str cove_folder: path to a folder with genome coverage by tracks
    :param str cove_prefix: prefix used in uniwig for creating coverage
    :param str chrom: chromosome to be processed
    :param str file_out: output file with the universe
    """
    model_lh.read_chrom(chrom)
    chrom_model = model_lh[chrom]
    start = read_chromosome_from_bw(os.path.join(cove_folder, f"{cove_prefix}_start.bw"), chrom)
    core = read_chromosome_from_bw(os.path.join(cove_folder, f"{cove_prefix}_core.bw"), chrom)
    end = read_chromosome_from_bw(os.path.join(cove_folder, f"{cove_prefix}_end.bw"), chrom)
    cove = np.zeros((len(start), 3), dtype=np.uint16)
    cove[:, 0] = start
    cove[:, 1] = core
    cove[:, 2] = end
    full_start, full_end = find_full(cove)
    path = np.full(len(cove), 3, dtype=np.uint8)
    for s, e in zip(full_start, full_end):
        res = process_part(
            cove[s:e],
            chrom_model["start"],
            chrom_model["core"],
            chrom_model["end"],
        )
        path[s:e] = res

    predictions_to_bed(path, chrom, file_out)


def ml_universe(model_file, cove_folder, cove_prefix, file_out):
    """
    Make ML flexible universe
    :param str model_file: input name with likelihood models
    :param str file_out: output file with the universe
    :param str cove_folder: path to a folder with genome coverage by tracks
    :param str cove_prefix: prefix used in uniwig for creating coverage
    """
    if os.path.isfile(file_out):
        raise Exception(f"File : {file_out} exists")
    lh_model = ModelLH(model_file)
    chroms = sorted(lh_model.chromosomes_list, key=cmp_to_key(natural_chr_sort))
    for C in chroms:
        make_ml_flexible_universe(lh_model, cove_folder, cove_prefix, C, file_out)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import torch.nn as nn

from ..io import RegionSet

if TYPE_CHECKING:
    from ..tokenization import Tokenizer


class Model(nn.Module):
    """Class representing an *actual* model, that is, weights, etc"""

    pass


class ExModel(ABC):
    """
    An Extended Model is a 3-part object consisting of a tokenizer (T), a
    universe/vocabulary (U), and a model (M). The tokenizer is used to tokenize
    region sets into the universe. The model is defined on the universe.
    """

    model: Model
    universe: RegionSet
    tokenizer: "Tokenizer"

    @abstractmethod
    def __init__(self, model_path: str = None, tokenizer: "Tokenizer" = None, device: str = None):
        """
        Initialize the model. If model_path is not None, load the model from
        huggingface.

        :param str model_path: Path to the model on huggingface to load
        :param Tokenizer tokenizer: Tokenizer to use
        :param str device: Device to use (e.g. "cpu", "cuda:0", etc)
        """
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import pyBigWig


def ana_region(region, start_s):
    """Helper for saving HMM prediction into a file"""
    start_e = start_s + np.where(region == 1)[0][0]
    end_s = start_s + np.where(region == 2)[0][0]
    return start_e, end_s


def predictions_to_bed(states, chrom, bedname, save_max_cove=False, cove_file=None):
    """
    Save HMM prediction into a file
    :param ndarray states: result of HMM prediction
    :param str chrom: which chromosome is being analyzed
    :param str bedname: path to the output file
    :param bool save_max_cove: whether to save the maximum peak coverage to output
     file, can result in nonstandard bed file
    :param str cove_file: file with core coverage, require for saving maximum peak coverage
    """
    ind = np.argwhere(states != 3)
    ind = ind.flatten()
    start_s = ind[0]
    to_file = []
    line = chrom + "\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n"
    if save_max_cove:
        coverage = pyBigWig.open(cove_file)
    for i in range(1, len(ind)):
        if ind[i] - ind[i - 1] != 1:
            end_e = ind[i - 1]
            region = states[start_s : end_e + 1]
            res = ana_region(region, start_s)
            save_start_e, save_end_s = res
            val = 0
            if save_max_cove:
                val = coverage.stats(chrom, int(start_s), int(end_e) + 1, type="max")
                val = int(val[0])
            to_file.append(
                line.format(
                    start_s,
                    end_e + 1,
                    "universe",
                    val,
                    ".",
                    save_start_e,
                    save_end_s,
                    "0,0,255",
                )
            )
            start_s = ind[i]
    if states[ind[-1]] == 2:
        region = states[start_s : ind[-1] + 1]
        res = ana_region(region, start_s)
        save_start_e, save_end_s = res
        val = 0
        if save_max_cove:
            val = coverage.stats(chrom, int(start_s), int(ind[-1]) + 1, type="max")
            val = int(val[0])
        to_file.append(
            line.format(
                start_s,
                ind[-1] + 1,
                "universe",
                val,
                ".",
                save_start_e,
                save_end_s,
                "0,0,255",
            )
        )
    with open(bedname, "a") as f:
        f.writelines(to_file)


def find_full_full_pos(seq, gap_size=1000, area_size=500):
    """Look for nonzero positions in coverage matrix, when most of the positions are zero
    :param ndarray seq: vector with information about non-zero positions
    :param int gap_size: size of minimum gap between non-zero positions that are separated
    :param int area_size: size of the area around non-zero positions to be included in the result
    :return list: list of starts of non-zero regions and list of ends of non-zero regions
    """
    size = len(seq)
    seq = np.argwhere(seq >= 1).flatten()
    starts, ends = [], []
    if seq[0] > gap_size:
        starts.append(int(seq[0] - area_size))
    else:
        starts.append(0)
    for e in range(1, len(seq)):
        if seq[e] - seq[e - 1] > gap_size:
            ends.append(int(seq[e - 1] + area_size))
            starts.append(int(seq[e] - area_size))
    ends.append(min(int(seq[-1] + area_size), size))
    return starts, ends


def find_full_empty_pos(seq, gap_size=10000, area_size=1000):
    """Look for nonzero positions in coverage matrix, when most of the positions are nonzero
    :param ndarray seq: vector with information about non-zero positions
    :param int gap_size: size of minimum gap between non-zero positions that are separated
    :param int area_size: size of the area around non-zero positions to be included in the result
    :return list: list of starts of non-zero regions and list of ends of non-zero regions
    """
    size = len(seq)
    seq = np.argwhere(seq == 0).flatten()
    starts, ends = [], []
    gap_len = 0
    gap_start = 0
    looking_for_first = True
    for e in range(1, len(seq)):
        if seq[e] - seq[e - 1] == 1:
            gap_len += 1
        else:
            if gap_len >= gap_size:
                starts.append(gap_start)
                ends.append(seq[e - 1])
                looking_for_first = False
            elif looking_for_first:
                starts.append(gap_start)
                ends.append(seq[e - 1])
                looking_for_first = False
            gap_len = 1
            gap_start = seq[e]
    starts_res = [max(0, i - area_size) for i in ends]
    end_res = [i + area_size for i in starts[1:]] + [size]
    if not starts_res:
        starts_res = [0]
    return starts_res, end_res


def find_full(seq):
    """Look for nonzero positions in coverage matrix"""
    seq = np.sum(seq, axis=1, dtype=np.uint8)
    full_pos_no = np.sum(seq >= 1)
    if full_pos_no < len(seq) - full_pos_no:
        return find_full_full_pos(seq)
    else:
        return find_full_empty_pos(seq)



from abc import ABC, abstractmethod
from typing import Dict, List, Union

import numpy as np


class EmSearchBackend(ABC):
    """
    An abstract class representing Embedding Search Backends. This allows
    backends to be either a qdrant server or a local in-memory NMS index, or
    anything, really. This allows us to use the same interface for both.
    """

    def __init__(self, embeddings: np.ndarray = None, labels: list = None) -> None:
        if embeddings:
            self.load(embeddings, labels)

    @abstractmethod
    def load(
        self,
        vectors: np.ndarray,
        ids: Union[np.ndarray, None] = None,
        payloads: Union[List[Dict[str, str]], None] = None,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def search(
        self,
        query: np.ndarray,
        limit: int,
        with_payload: bool = True,
        with_vectors: bool = True,
        offset: int = 0,
    ) -> List[Dict]:
        """
        Search for the nearest neighbors of the given embedding

        :param query: the embedding to search for
        :param k: the number of results to return
        :return: a list of (id, score) pairs
        """
        raise NotImplementedError()

    @abstractmethod
    def __len__(self) -> int:
        """
        Return the number of embeddings in the backend
        """
        raise NotImplementedError()

    @abstractmethod
    def retrieve_info(self, key) -> List[Dict]:
        """
        with a list of storage ids, return matching vectors and their information
        """
        raise NotImplementedError()



from time import time
from typing import Dict, List, Optional

import numpy as np
import pyBigWig


def natural_chr_sort(a, b):
    ac = a.replace("chr", "")
    ac = ac.split("_")[0]
    bc = b.replace("chr", "")
    bc = bc.split("_")[0]
    if bc.isnumeric() and ac.isnumeric() and bc != ac:
        if int(bc) < int(ac):
            return 1
        elif int(bc) > int(ac):
            return -1
        else:
            return 0
    else:
        if b < a:
            return 1
        elif a < b:
            return -1
        else:
            return 0


def timer_func(func):
    def wrap_func(*args, **kwargs):
        t1 = time()
        result = func(*args, **kwargs)
        t2 = time()
        print(f"Function {func.__name__!r} executed in {(t2-t1)/60:.4f}min")
        return result

    return wrap_func


def read_chromosome_from_bw(file, chrom):
    bw = pyBigWig.open(file)
    chrom_size = bw.chroms(chrom)
    if pyBigWig.numpy:
        cove = bw.values(chrom, 0, chrom_size, numpy=True)
    else:
        cove = bw.values(chrom, 0, chrom_size)
        cove = np.array(cove)
    cove[np.isnan(cove)] = 0
    return cove.astype(np.uint16)


def find_path(hierarchy: Dict[str, Dict], path: List[str], cell_type: str) -> Optional[List[str]]:
    """
    Find the path from the root to a given cell type in a hierarchy.

    :param hierarchy: A dictionary representing the hierarchy.
    :param path: The current path.
    :param cell_type: The cell type to find.

    :return: The path from the root to the cell type. (a list of strings, ... or None)
    """
    if cell_type in hierarchy:
        return path + [cell_type]

    for key in hierarchy:
        sub_path = find_path(hierarchy[key], path + [key], cell_type)
        if sub_path:
            return sub_path

    return None


def find_lca(path1: List[str], path2: List[str]) -> int:
    """
    Find the lowest common ancestor (LCA) of two paths.

    :param path1: The first path.
    :param path2: The second path.
    """
    min_length = min(len(path1), len(path2))
    for i in range(min_length):
        if path1[i] != path2[i]:
            return i - 1
    return min_length - 1


def compute_cell_hierarchy_distance(
    hierarchy: Dict[str, Dict], cell1: str, cell2: str
) -> Optional[int]:
    """
    Compute the distance between two cell types in a hierarchy.

    The distance is the number of edges between the two cells in the hierarchy.

    :param hierarchy: A dictionary representing the hierarchy.
    :param cell1: The first cell type.
    :param cell2: The second cell type.

    :return: The distance between the two cell types. (an integer, ... or None)
    """
    # Find paths from root to both cells
    path1 = find_path(hierarchy, [], cell1)
    path2 = find_path(hierarchy, [], cell2)

    if not path1 or not path2:
        return None  # One of the cells doesn't exist in the hierarchy

    # Find the lowest common ancestor (LCA)
    lca_index = find_lca(path1, path2)

    # Distance is the sum of the lengths from LCA to both nodes
    distance = (len(path1) - lca_index - 1) + (len(path2) - lca_index - 1)

    return distance



import logging
import pickle
from typing import Dict, List, Set, Tuple, Union

import numpy as np

from ..const import PKG_NAME
from .backends.filebackend import HNSWBackend

_LOGGER = logging.getLogger(PKG_NAME)


def merge_backends(
    backends_to_merge: List[HNSWBackend], local_index_path: str, dim: int
) -> HNSWBackend:
    """
    Merge multiple backends into one

    :param backends_to_merge: a list of [HNSWBackend]
    :param local_index_path: the path to the local index file of the merged output HNSWBackend
    :param dim: the dimension of vectors stored in the HNSWBackend

    :return: a HNSWBackend that comes from merge all HNSWBackend in the input list backends_to_merge
    """

    result_backend = HNSWBackend(
        local_index_path=local_index_path,
        payloads={},
        dim=dim,
    )

    for backend in backends_to_merge:
        result_vecs = []
        result_payloads = []
        for j in range(len(backend)):
            result_vecs.append(backend.idx.get_items([j], return_type="numpy")[0])
            result_payloads.append(backend.payloads[j])

        result_backend.load(vectors=np.array(result_vecs), payloads=result_payloads)

    return result_backend



import os

from setuptools import setup

PACKAGE_NAME = "geniml"

# Ordinary dependencies
DEPENDENCIES = []
with open("requirements/requirements-basic.txt", "r") as reqs_file:
    for line in reqs_file:
        if not line.strip():
            continue
        DEPENDENCIES.append(line)

# Additional keyword arguments for setup()
extra = {"install_requires": DEPENDENCIES}

with open(PACKAGE_NAME + "/_version.py", "r") as versionfile:
    version = versionfile.readline().split()[-1].strip("\"'\n")

# Optional dependencies
# Extras requires a dictionary and not a list?
with open("requirements/requirements-ml.txt", "r") as reqs_file:
    ml_dep = []
    for line in reqs_file:
        if not line.strip():
            continue
        ml_dep.append(line.strip())

with open("requirements/requirements-test.txt", "r") as reqs_file:
    test_dep = []
    for line in reqs_file:
        if not line.strip():
            continue
        test_dep.append(line.strip())

extra["install_requires"] = DEPENDENCIES
extra["extras_require"] = {
    "ml": ml_dep,
    "test": test_dep,
}


with open("README.md") as f:
    long_description = f.read()

setup(
    name=PACKAGE_NAME,
    packages=[
        PACKAGE_NAME,
        "geniml.atacformer",
        "geniml.assess",
        "geniml.bedspace",
        "geniml.bedshift",
        "geniml.eval",
        "geniml.likelihood",
        "geniml.models",
        "geniml.region2vec",
        "geniml.scembed",
        "geniml.tokenization",
        "geniml.universe",
        "geniml.io",
        "geniml.text2bednn",
        "geniml.bbclient",
        "geniml.search",
        "geniml.search.backends",
        "geniml.search.interfaces",
        "geniml.search.query2vec",
        "geniml.nn",
    ],
    version=version,
    long_description=long_description,
    long_description_content_type="text/markdown",
    description="Genomic interval toolkit",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    license="BSD2",
    entry_points={
        "console_scripts": [
            "geniml = geniml.cli:main",
            "bedshift = geniml.bedshift.bedshift:main",
        ],
    },
    keywords="bioinformatics, sequencing, ngs",
    package_data={"geniml": [os.path.join("geniml", "*")]},
    include_package_data=True,
    url="https://docs.bedbase.org/geniml/",
    author="Nathan Sheffield",
    **extra,
)



from typing import Literal

PAD_CHR = "chrPAD"
PAD_START = 0
PAD_END = 0

MASK_CHR = "chrMASK"
MASK_START = 0
MASK_END = 0

DEFAULT_EMBEDDING_DIM = 768
DEFAULT_CONTEXT_SIZE = 2048

MASK_RATE = 0.15
REPLACE_WITH_MASK_RATE = 0.8
REPLACE_WITH_RANDOM_RATE = 0.1
KEEP_RATE = 0.1

CONFIG_FILE_NAME = "config.yaml"
MODEL_FILE_NAME = "checkpoint.pt"
UNIVERSE_FILE_NAME = "universe.bed"
UNIVERSE_CONFIG_FILE_NAME = "universe.toml"

POOLING_TYPES = Literal["mean", "max"]
POOLING_METHOD_KEY = "pooling_method"
D_MODEL_KEY = "d_model"
VOCAB_SIZE_KEY = "vocab_size"
N_HEADS_KEY = "n_heads"
N_LAYERS_KEY = "n_layers"
CONTEXT_SIZE_KEY = "context_size"
D_FF_KEY = "d_ff"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
from functools import cmp_to_key

import numpy as np
import pyBigWig

from geniml.utils import natural_chr_sort, timer_func


def get_uni(file, chrom, cutoff=None):
    """For each position check if coverage is bigger than cut-off;
    if cut-off not provided calculate value that gives
    maximum likelihood universe
    :param str file: coverage file
    :param str chrom: chromosome to analyse
    :param int cutoff: base pairs with values grater of equal to cut-off can be included in universe
    :return ndarray: vector with universes states; 0 - background, 1- universe"""
    file = pyBigWig.open(file)
    if pyBigWig.numpy:
        track = file.values(chrom, 0, file.chroms(chrom), numpy=True)
    else:
        track = file.values(chrom, 0, file.chroms(chrom))
        track = np.array(track)
    track[np.isnan(track)] = 0
    track = track.astype(np.uint16)
    if cutoff is None:
        cutoff = np.sum(track) / len(track)
    inter_pos = track >= cutoff
    file.close()
    return inter_pos


def save_simple(file_out, inter_pos, chrom):
    """
    Save cut-off universe to a file without any processing
    :param str file_out: output file
    :param bool vector inter_pos: whether each position should be included in universe
    :param str chrom: chromosome to analyse
    """
    inter_pos_uni = np.argwhere(inter_pos)
    start = inter_pos_uni[0][0]
    with open(file_out, "a") as f:
        for i in range(1, len(inter_pos_uni)):
            if inter_pos_uni[i] - inter_pos_uni[i - 1] != 1:
                end = inter_pos_uni[i - 1][0] + 1
                f.write(f"{chrom}\t{start}\t{end}\n")
                start = inter_pos_uni[i][0]
        end = inter_pos_uni[-1][0] + 1
        f.write(f"{chrom}\t{start}\t{end}\n")


def marge_filter(file_out, inter_pos, chrom, merge_dist=100, size_flt=1000):
    """
    Save cut-off universe to a file with filtering region size and merging close regions
    :param file_out: output file
    :param bool vector inter_pos: whether each position should be included in universe
    :param str chrom: chromosome to analyse
    :param int merge_dist: regions closer than merge_dist will be merged into one
    :param int size_flt: regions smaller than size_flt will not be reported
    """
    inter_pos_uni = np.argwhere(inter_pos)
    start = inter_pos_uni[0][0]
    with open(file_out, "a") as f:
        for i in range(1, len(inter_pos_uni)):
            if inter_pos_uni[i] - inter_pos_uni[i - 1] >= merge_dist:
                end = inter_pos_uni[i - 1][0] + 1
                if end - start >= size_flt:
                    f.write(f"{chrom}\t{start}\t{end}\n")
                start = inter_pos_uni[i][0]
        end = inter_pos_uni[-1][0] + 1
        if end - start >= size_flt:
            f.write(f"{chrom}\t{start}\t{end}\n")


def cc_universe(cove, file_out, cove_prefix="all", merge=0, filter_size=0, cutoff=None):
    """
    Create cut-off coverage universe based on coverage track
    :param str cove: path to coverage folder
    :param str cove_prefix: prefix of the coverage file
    :param int merge: regions closer than this value will be merged into one
    :param int filter_size: regions smaller than this value will not be reported
    :param str file_out: output file
    :param int cutoff: base pairs with coverage equal to or greater than this value will be included in the universe
    """
    if os.path.isfile(file_out):
        raise Exception(f"File : {file_out} exists")
    file = os.path.join(cove, f"{cove_prefix}_core.bw")
    bw_start = pyBigWig.open(file)
    chroms = bw_start.chroms()
    bw_start.close()
    chroms_key = list(chroms.keys())
    chroms_key = sorted(chroms_key, key=cmp_to_key(natural_chr_sort))
    chroms = {i: chroms[i] for i in chroms_key}
    for chrom in chroms:
        if chroms[chrom] > 0:
            inter_pos = get_uni(file, chrom, cutoff)
            if merge == 0 and filter_size == 0:
                save_simple(file_out, inter_pos, chrom)
            else:
                marge_filter(file_out, inter_pos, chrom, merge, filter_size)



import os
import random
from typing import Dict, List

import numpy as np
import pytest

from geniml.io import RegionSet
from geniml.region2vec.main import Region2VecExModel
from geniml.search import BED2BEDSearchInterface, BED2Vec, Text2BEDSearchInterface, Text2Vec
from geniml.search.backends import BiVectorBackend, HNSWBackend, QdrantBackend
from geniml.search.backends.filebackend import DEP_HNSWLIB
from geniml.search.interfaces.mlfree import BiVectorSearchInterface

DATA_FOLDER_PATH = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "tests", "data"
)

random.seed(100)
np.random.seed(100)


@pytest.fixture
def bed_folder():
    """
    folder where testing bed files are stored
    """

    return os.path.join(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
        "tests",
        "data",
        "hg38_sample",
    )


@pytest.fixture
def filenames(bed_folder):
    """
    list of bed file names
    """

    return [
        "ENCX3P",
        "ENCN4Z",
        "ENC7VQ",
        "ENCY6R",
        "ENCJ9K",
        "ENCD8T",
        "ENCQ1A",
        "ENCM2F",
        "ENCKMR",
    ]


@pytest.fixture
def metadata():
    """
    mock metadata for testing
    """

    return {
        "ENCX3P": {"biosample": "HEK293", "target": "H3K27ac", "organ": ["kidney", "epithelium"]},
        "ENCN4Z": {"biosample": "HEK293", "target": "CTCF", "organ": ["kidney"]},
        "ENC7VQ": {"biosample": "HEK293", "target": "TBP", "organ": ["kidney", "epithelium"]},
        "ENCY6R": {"biosample": "A549", "target": "H3K27ac", "organ": ["epithelium", "lung"]},
        "ENCJ9K": {"biosample": "A549", "target": "CTCF", "organ": ["lung"]},
        "ENCD8T": {"biosample": "K562", "target": "TBP", "organ": ["blood"]},
        "ENCQ1A": {"biosample": "K562", "target": "H3K27ac", "organ": ["blood"]},
        "ENCM2F": {"biosample": "K562", "target": "CTCF", "organ": ["blood"]},
        "ENCKMR": {"biosample": "apple"},
    }


@pytest.fixture
def annotations():
    return [
        "HEK293",
        "A549",
        "K562",
        "H3K27ac",
        "CTCF",
        "TBP",
        "kidney",
        "epithelium",
        "lung",
        "blood",
        "apple",
    ]


@pytest.fixture
def annotation_matches():
    return {
        "HEK293": [0, 1, 2],
        "A549": [3, 4],
        "K562": [5, 6, 7],
        "H3K27ac": [0, 3, 5],
        "CTCF": [1, 4, 7],
        "TBP": [2, 6],
        "kidney": [0, 2],
        "epithelium": [0, 2, 3],
        "lung": [3, 4],
        "blood": [5, 6, 7],
        "apple": [8],
    }


@pytest.fixture
def uuids():
    return [
        "7bbab414-053d-4c06-9085-d3ca894dc8b8",
        "8b3fa142-8866-4b4c-9df8-7734b4ef9f2a",
        "478c9b96-3b4c-41c3-af56-68e8c39de0a3",
        "971c58a5-c126-433b-887c-4184184cbce6",
        "b28381f8-82ce-4b19-86c0-34a2d368e3b3",
        "d6f1060e-6e14-4faf-8711-f25ed5c6618e",
        "920ef6f6-f821-46f9-9d11-3516119feeec",
        "6ec9d4a4-a481-43dc-81f3-098953c77b0a",
        "ce5345d8-84a1-4c6c-9427-145a3f207805",
    ]


@pytest.fixture
def bed_payloads(filenames, metadata):
    """
    mock list of label dictionaries for testing
    """

    output_list = []
    for name in filenames:
        output_list.append({"name": name, "metadata": metadata[name]})
    return output_list


@pytest.fixture
def metadata_payloads(annotations, annotation_matches):
    """
    mock list of label dictionaries for testing
    """

    output_list = []
    for tag in annotations:
        output_list.append({"text": tag, "matched_files": annotation_matches[tag]})
    return output_list


@pytest.fixture
def bed_collection():
    """
    Returns: bed_collection name for qdrant client storage
    """

    return "hg38_sample"


@pytest.fixture
def metadata_collection():
    """
    Returns: bed_collection name for qdrant client storage
    """

    return "bed_metadata"


@pytest.fixture
def bed_embeddings(filenames):
    """
    mock embedding vectors for testing
    """

    return np.random.random((len(filenames), 100))


@pytest.fixture
def text_embeddings(annotations):
    """
    mock embedding vectors for testing
    """

    return np.random.random((len(annotations), 384))


@pytest.fixture
def int_ids(filenames):
    """
    list of randomly sampled integer_ids
    """
    return random.sample(range(len(filenames)), 3)


@pytest.fixture
def ids(uuids):
    ids_with_hyphen = random.sample(uuids, 3)
    return [uuid.replace("-", "") for uuid in ids_with_hyphen]


# @pytest.fixture
@pytest.fixture(scope="module")
def temp_data_dir(tmp_path_factory):
    # temporal index folder
    return tmp_path_factory.mktemp("data")


@pytest.fixture(scope="module")
def temp_bed_idx_path(temp_data_dir):
    # temporal index path
    return temp_data_dir / "testing_idx.bin"


@pytest.fixture(scope="module")
def temp_metadata_idx_path(temp_data_dir):
    # temporal index path
    return temp_data_dir / "testing_metadata_idx.bin"


@pytest.fixture(scope="module")
def bed_hnswb(temp_bed_idx_path):
    # init backend
    return HNSWBackend(local_index_path=str(temp_bed_idx_path))


@pytest.fixture(scope="module")
def metadata_hnswb(temp_metadata_idx_path):
    # init backend
    return HNSWBackend(local_index_path=str(temp_metadata_idx_path), dim=384)


@pytest.fixture
def bed_folder():
    """
    :return: the path to the folder where testing bed files are stored
    """
    return os.path.join(DATA_FOLDER_PATH, "hg38_sample")


@pytest.fixture
def r2v_hf_repo():
    """
    :return: the huggingface repo of region2vec model
    """
    return "databio/r2v-encode-hg38"


@pytest.fixture
def v2v_hf_repo():
    """
    Returns: the huggingface repo of vec2vec model
    """
    return "databio/v2v-sentencetransformers-encode"


@pytest.fixture
def query_term():
    """
    Returns: a query string
    """
    return "human, kidney, blood"


@pytest.fixture
def nl_embed_repo():
    """
    Returns: text embedding model for search backend
    """
    return "sentence-transformers/all-MiniLM-L6-v2"


@pytest.fixture
def query_bed():
    """
    Returns: path to a BED file in testing data
    """
    return "./data/s1_a.bed"


def cosine_similarity(vec1: np.array, vec2: np.array) -> float:
    # Ensure the vectors have shape (100,)
    assert vec1.shape == (100,) and vec2.shape == (100,)

    # Compute the dot product of the two vectors
    dot_product = np.dot(vec1, vec2)

    # Compute the magnitude (L2 norm) of each vector
    magnitude_vec1 = np.linalg.norm(vec1)
    magnitude_vec2 = np.linalg.norm(vec2)

    # Compute the cosine similarity
    if magnitude_vec1 == 0 or magnitude_vec2 == 0:
        return 0.0  # Avoid division by zero

    cosine_sim = dot_product / (magnitude_vec1 * magnitude_vec2)

    return cosine_sim


@pytest.mark.skipif(
    "not config.getoption('--qdrant')",
    reason="Only run when --qdrant is given",
)
def test_QdrantBackend(filenames, bed_embeddings, bed_payloads, bed_collection, ids, uuids):
    def search_results_test(search_results):
        assert isinstance(search_results, list)
        for result in search_results:
            assert isinstance(result, dict)  # only target pairs
            assert isinstance(result["id"], str)
            assert isinstance(result["score"], float)

            assert isinstance(result["vector"], list)
            for i in result["vector"]:
                assert isinstance(i, float)
            assert isinstance(result["payload"], dict)
            assert isinstance(result["payload"]["name"], str)
            assert isinstance(result["payload"]["metadata"], dict)

    qd_search_backend = QdrantBackend(collection=bed_collection)
    # load data
    qd_search_backend.load(bed_embeddings, payloads=bed_payloads, ids=uuids)
    # test searching
    query_vec = np.random.random(
        100,
    )
    search_results = qd_search_backend.search(
        query_vec,
        5,
        with_payload=True,
        with_vectors=True,
    )

    search_results_test(search_results)

    assert len(qd_search_backend) == len(filenames)

    # test information retrieval
    retrieval_results = qd_search_backend.retrieve_info(ids, True)
    assert len(retrieval_results) == len(ids)
    assert isinstance(retrieval_results, list)
    for i in range(len(ids)):
        assert ids[i] == retrieval_results[i]["id"]

        client_retrieval = qd_search_backend.qd_client.retrieve(
            bed_collection, [ids[i]], with_vectors=True
        )

        assert retrieval_results[i]["vector"] == client_retrieval[0].vector
        assert retrieval_results[i]["payload"] == client_retrieval[0].payload

    # test batch search
    batch_query = np.random.random((6, 100))
    batch_result = qd_search_backend.search(
        batch_query, limit=3, with_payload=True, with_vectors=True
    )
    assert len(batch_result) == 6
    for batch in batch_result:
        search_results_test(batch)

    qd_search_backend.qd_client.delete_collection(qd_search_backend.collection)


@pytest.mark.skipif(
    DEP_HNSWLIB == False, reason="This test require installation of hnswlib (optional)"
)
def test_HNSWBackend_load(filenames, bed_embeddings, bed_payloads, bed_hnswb, ids):
    num_upload = len(filenames)

    # batches to load
    labels_1 = bed_payloads[: num_upload // 2]
    labels_2 = bed_payloads[num_upload // 2 :]
    embeddings_1 = bed_embeddings[: num_upload // 2]
    embeddings_2 = bed_embeddings[num_upload // 2 :]

    # load first batch
    bed_hnswb.load(embeddings_1, payloads=labels_1)
    assert len(bed_hnswb) == num_upload // 2

    # load second batch
    bed_hnswb.load(embeddings_2, payloads=labels_2)
    assert len(bed_hnswb) == num_upload


@pytest.mark.skipif(
    DEP_HNSWLIB == False, reason="This test require installation of hnswlib (optional)"
)
@pytest.mark.dependency(depends=["test_HNSWBackend_load"])
def test_HNSWBackend_search(filenames, bed_hnswb, int_ids):
    def search_result_check(dict_list: List[Dict], backend: HNSWBackend, with_dist: bool = False):
        """
        repeated test of the output of search / retrieve_info function of HNSWBackend to check if the result matches the content in index

        :param dict_list: the result, which is supposed to be a list of dictionary
        :param backend: the HNSWBackend to be tested
        :param with_dist: whether distance score is included in the result
        :return:
        """
        index = backend.idx
        assert isinstance(dict_list, list)

        for result in dict_list:
            assert isinstance(result, dict)
            assert isinstance(result["id"], int)
            if with_dist:
                assert isinstance(result["distance"], float)
            assert isinstance(result["payload"], dict)
            assert isinstance(result["vector"], np.ndarray)

            assert (
                result["vector"] == index.get_items([result["id"]], return_type="numpy")[0]
            ).all()
            for num in result["vector"]:
                assert isinstance(num, np.float32)

    # bed_hnswb = pytestconfig.cache.get('shared_backend', None)
    assert len(bed_hnswb) == len(filenames)
    # test searching with one vector (np.ndarray with shape (dim,))
    query_vec = np.random.random(
        100,
    )
    single_vec_search = bed_hnswb.search(
        query_vec,
        3,
    )

    single_vec_search_offset = bed_hnswb.search(
        query_vec,
        3,
        offset=2,
    )

    for j in range(len(single_vec_search)):
        assert single_vec_search_offset[j]["id"] == single_vec_search[j]["id"]
        assert single_vec_search_offset[j]["distance"] == single_vec_search[j]["distance"]
        assert (
            single_vec_search_offset[j]["payload"]["metadata"]
            == single_vec_search[j]["payload"]["metadata"]
        )
    search_result_check(single_vec_search, bed_hnswb, True)
    search_result_check(single_vec_search_offset, bed_hnswb, True)

    # test searching with multiple vectors (np.ndarray with shape (n, dim))
    multiple_vecs_search = bed_hnswb.search(np.random.random((7, 100)), 5)
    assert isinstance(multiple_vecs_search, list)
    assert len(multiple_vecs_search) == 7
    for i in range(len(multiple_vecs_search)):
        search_result_check(multiple_vecs_search[i], bed_hnswb, True)

    # test information retrieval / get items
    retrieval_results = bed_hnswb.retrieve_info(int_ids, True)
    search_result_check(retrieval_results, bed_hnswb, False)


@pytest.mark.skipif(
    DEP_HNSWLIB == False, reason="This test require installation of hnswlib (optional)"
)
@pytest.mark.dependency(depends=["test_HNSWBackend_load"])
def test_HNSWBackend_save(filenames, bed_hnswb, bed_embeddings, temp_bed_idx_path, temp_data_dir):
    # test saving from local
    new_hnswb = HNSWBackend(local_index_path=str(temp_bed_idx_path), payloads=bed_hnswb.payloads)
    assert new_hnswb.idx.max_elements == bed_embeddings.shape[0]

    for i in range(bed_embeddings.shape[0]):
        old_result = bed_hnswb.idx.get_items([i], return_type="numpy")
        new_result = new_hnswb.idx.get_items([i], return_type="numpy")
        assert (old_result == new_result).all()

    # test a bug:
    new_idx_path = temp_data_dir / "new_idx.bin"
    empty_hnswb = HNSWBackend(local_index_path=str(new_idx_path))
    assert len(empty_hnswb.payloads) == 0


@pytest.mark.skipif(
    DEP_HNSWLIB == False, reason="This test require installation of hnswlib (optional)"
)
@pytest.mark.dependency(depends=["test_HNSWBackend_load"])
@pytest.mark.skipif(
    "not config.getoption('--qdrant')",
    reason="Only run when --qdrant is given",
)
@pytest.mark.skipif(
    "not config.getoption('--huggingface')",
    reason="Only run when --huggingface is given",
)
def test_BiVectorBackend(
    bed_hnswb,
    metadata_hnswb,
    bed_collection,
    bed_embeddings,
    bed_payloads,
    metadata_collection,
    text_embeddings,
    metadata_payloads,
    nl_embed_repo,
):
    def search_result_test(search_results: Dict, rank: bool):
        min_score = 100.0
        max_rank = -1
        for result in search_results:
            assert isinstance(result, dict)  # only target pairs
            assert isinstance(result["id"], int)

            if not rank:
                assert isinstance(result["score"], float)
                assert 0 <= result["score"] <= 1
                assert result["score"] <= min_score
                min_score = result["score"]
            else:
                assert isinstance(result["max_rank"], int)
                assert result["max_rank"] >= max_rank
                max_rank = result["max_rank"]

            assert isinstance(result["vector"], list) or isinstance(result["vector"], np.ndarray)
            if isinstance(result["vector"], list):
                for i in result["vector"]:
                    assert isinstance(i, float)
            assert isinstance(result["payload"], dict)
            assert isinstance(result["payload"]["name"], str)
            assert isinstance(result["payload"]["metadata"], dict)

    def bivec_test(bivec_backend, dist: bool = False, rank: bool = False):
        query_vec = np.random.random(
            384,
        )
        search_results = bivec_backend.search(
            query_vec, 2, with_payload=True, with_vectors=True, distance=dist, rank=rank
        )
        assert isinstance(search_results, list)
        assert len(search_results) == 2

        search_result_test(search_results, rank)

    # test QdrantBackend
    bed_backend = QdrantBackend(collection=bed_collection)
    # load data
    bed_backend.load(bed_embeddings, payloads=bed_payloads)

    text_backend = QdrantBackend(collection=metadata_collection, dim=384)
    text_backend.load(text_embeddings, payloads=metadata_payloads)

    bivec_qd_backend = BiVectorBackend(text_backend, bed_backend)
    bivec_test(bivec_qd_backend, rank=True)
    bivec_test(bivec_qd_backend, rank=False)
    # test QdrantBackend + Interface
    bivec_qd_interface = BiVectorSearchInterface(bivec_qd_backend, nl_embed_repo)
    interface_result = bivec_qd_interface.query_search(
        "lung cancer cell line", 2, with_payload=True, with_vectors=True, rank=True
    )

    search_result_test(interface_result, rank=True)

    interface_result = bivec_qd_interface.query_search(
        "lung cancer cell line", 2, with_payload=True, with_vectors=True, rank=False
    )

    search_result_test(interface_result, rank=False)

    bivec_qd_backend.metadata_backend.qd_client.delete_collection(text_backend.collection)
    bivec_qd_backend.bed_backend.qd_client.delete_collection(bed_backend.collection)

    # test HNSWBackend
    metadata_hnswb.load(text_embeddings, payloads=metadata_payloads)
    bivec_hnsw_backend = BiVectorBackend(metadata_hnswb, bed_hnswb)
    bivec_test(bivec_hnsw_backend, dist=True, rank=True)
    bivec_test(bivec_hnsw_backend, dist=True, rank=False)
    # test HNSWBackend + Interface
    bivec_hnsw_interface = BiVectorSearchInterface(bivec_hnsw_backend, nl_embed_repo)
    interface_result = bivec_hnsw_interface.query_search(
        "lung cancer cell line", 2, with_payload=True, with_vectors=True, distance=True, rank=True
    )

    search_result_test(interface_result, rank=True)

    interface_result = bivec_hnsw_interface.query_search(
        "lung cancer cell line", 2, with_payload=True, with_vectors=True, distance=True, rank=False
    )

    search_result_test(interface_result, rank=False)


@pytest.mark.skipif(
    "not config.getoption('--huggingface')",
    reason="Only run when --huggingface is given",
)
def test_bed2vec(r2v_hf_repo, bed_folder):
    bed2vec = BED2Vec(r2v_hf_repo)
    for bed_name in os.listdir(bed_folder):
        assert bed2vec.forward(os.path.join(bed_folder, bed_name)).shape == (100,)


@pytest.mark.skipif(
    "not config.getoption('--huggingface')",
    reason="Only run when --huggingface is given",
)
def test_text2vec(nl_embed_repo, v2v_hf_repo):
    text2vec = Text2Vec(nl_embed_repo, v2v_hf_repo)
    assert text2vec.forward("Hematopoietic cells").shape == (100,)


@pytest.mark.skipif(
    "not config.getoption('--huggingface')",
    reason="Only run when --huggingface is given",
)
@pytest.mark.skipif(
    "not config.getoption('--qdrant')",
    reason="Only run when --qdrant is given",
)
def test_text2bed_search_interface(
    bed_folder,
    r2v_hf_repo,
    nl_embed_repo,
    v2v_hf_repo,
    bed_collection,
    query_term,
    tmp_path_factory,
):
    r2v_model = Region2VecExModel(r2v_hf_repo)
    query_dict = {"Mock1": [2, 3], "Mock2": [1], "Mock3": [2, 4, 5], "Mock4": [0]}
    vecs = []
    payloads = []

    for bed_name in os.listdir(bed_folder):
        rs = RegionSet(os.path.join(bed_folder, bed_name))
        # bed_vecs.append(r2v.encode(bed_path))
        region_embeddings = r2v_model.encode(rs)
        bed_file_embedding = np.mean(region_embeddings, axis=0)
        vecs.append(bed_file_embedding)
        file_payload = {"name": bed_name}
        payloads.append(file_payload)

    vecs = np.array(vecs)

    qd_search_backend = QdrantBackend(collection=bed_collection)
    qd_search_backend.load(vectors=vecs, payloads=payloads)
    # #
    text2vec = Text2Vec(nl_embed_repo, v2v_hf_repo)
    # #
    # # construct a search interface
    db_interface = Text2BEDSearchInterface(qd_search_backend, text2vec)
    db_search_result = db_interface.query_search(query_term, 5, offset=0)
    for i in range(len(db_search_result)):
        assert isinstance(db_search_result[i], dict)

    eval_results = db_interface.eval(query_dict)
    assert eval_results["Mean Average Precision"] > 0
    assert eval_results["Mean AUC-ROC"] > 0
    assert eval_results["Average R-Precision"] > 0

    # delete testing bed_collection
    db_interface.backend.qd_client.delete_collection(collection_name=bed_collection)

    # construct a search interface with file backend
    temp_data_dir = tmp_path_factory.mktemp("data")
    temp_idx_path = temp_data_dir / "testing_idx.bin"
    hnsw_backend = HNSWBackend(local_index_path=str(temp_idx_path))
    hnsw_backend.load(vectors=vecs, payloads=payloads)
    file_interface = Text2BEDSearchInterface(hnsw_backend, text2vec)

    file_search_result = file_interface.query_search(query_term, 5)
    for i in range(len(file_search_result)):
        assert isinstance(file_search_result[i], dict)

    # test evaluation
    eval_results = file_interface.eval(query_dict)
    assert eval_results["Mean Average Precision"] > 0
    assert eval_results["Mean AUC-ROC"] > 0
    assert eval_results["Average R-Precision"] > 0


@pytest.mark.skipif(
    "not config.getoption('--huggingface')",
    reason="Only run when --huggingface is given",
)
@pytest.mark.skipif(
    "not config.getoption('--qdrant')",
    reason="Only run when --qdrant is given",
)
def test_bed2bed_search_interface(
    bed_folder,
    r2v_hf_repo,
    bed_collection,
    query_bed,
    tmp_path_factory,
):
    r2v_model = Region2VecExModel(r2v_hf_repo)
    vecs = []
    payloads = []

    for bed_name in os.listdir(bed_folder):
        rs = RegionSet(os.path.join(bed_folder, bed_name))
        region_embeddings = r2v_model.encode(rs)
        bed_file_embedding = np.mean(region_embeddings, axis=0)
        vecs.append(bed_file_embedding)
        file_payload = {"name": bed_name}
        payloads.append(file_payload)

    vecs = np.array(vecs)

    qd_search_backend = QdrantBackend(collection=bed_collection)
    qd_search_backend.load(vectors=vecs, payloads=payloads)

    bed2vec = BED2Vec(r2v_hf_repo)

    # # construct a search interface
    db_interface = BED2BEDSearchInterface(qd_search_backend, bed2vec)
    db_search_result = db_interface.query_search(query_bed, 5, offset=0)
    for i in range(len(db_search_result)):
        assert isinstance(db_search_result[i], dict)

    # delete testing bed_collection
    db_interface.backend.qd_client.delete_collection(collection_name=bed_collection)

    # construct a search interface with file backend
    temp_data_dir = tmp_path_factory.mktemp("data")
    temp_idx_path = temp_data_dir / "testing_idx.bin"
    hnsw_backend = HNSWBackend(local_index_path=str(temp_idx_path))
    hnsw_backend.load(vectors=vecs, payloads=payloads)
    file_interface = Text2BEDSearchInterface(hnsw_backend, bed2vec)

    file_search_result = file_interface.query_search(query_bed, 5)
    for i in range(len(file_search_result)):
        assert isinstance(file_search_result[i], dict)
